/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <stdio.h>
#include <stdlib.h>

#include <cache.h>
#include <barrier.h>
#include <tsc.h>
#include <log.h>

/* Various matrices needed, static allocation! */

/* The big data matrix [A||B]. */
#pragma DATA_SECTION(M, "matrix")
double M[MATRIX_SIZE*MATRIX_SIZE];

/* The solution vector. */
#pragma DATA_SECTION(x, "matrix")
double x[MATRIX_SIZE];

/******************************************************************************/

struct param_t {
  long size;
  long repetitions;
};

const struct param_t parameters[] = {
#if 0
  {   127, 260000 },
  {   255,  62000 },
  {   511,  15000 },
  {  1023,   3200 },
  {  2047,    660 },
  {  4095,    130 },
  {  8159,     21 },
#endif
#if 0
  {   127, 1 },
  {   255, 1 },
  {   511, 1 },
  {  1023, 1 },
  {  2047, 1 },
  {  4095, 1 },
  {  8063, 1 },
  { -1, -1 },
#endif
  { MATRIX_SIZE-1, 1},
  { -1, -1},
  { 9*PANEL_SIZE-1, 1},
  {    -1,     -1 }
};

/******************************************************************************/

int
main(int argc, char **argv)
{
  size_t num_equations = 9*PANEL_SIZE - 1 /* MATRIX_SIZE - 1 */;
  uint64_t random_seed = 1234;
  size_t i;
  int ret;
  unsigned my_rank;
  unsigned num_ranks;
  uint64_t start_t, end_t;
  double time, flop;
  const struct param_t *cur_run = parameters;

  my_rank = cache_get_mm_id();
  num_ranks = cache_get_mm_num();
  while (cur_run->size > 0) {
    num_equations = cur_run->size;
    ++cur_run;
    barrier();

    cache_l1d_set_size(4);	/* 32kB L1D cache, 0kB SRAM*/
    cache_l2_set_size(0);		/* 0kB L2, 512kB SRAM */

    if (my_rank == 0) {
      log_write_name("linpack");
      log_write_param_uint64("size", num_equations);
      log_write_param_uint64("iterations", 1);
      log_write_param_uint64("ranks", num_ranks);
    }
    barrier();

    setup_m_panel();

    if (my_rank == 0) {
      init_problem(M, num_equations, random_seed);
      for(i=0; i<num_equations; ++i) x[i] = 0.0;
    }
    coherence_flush_all();
    barrier();
  
    if (my_rank == 0) log_write_start();
    start_t = tsc_get();
  
    solve(x, M, num_equations);

    coherence_flush_all();
    barrier();

    end_t = tsc_get();
    if (my_rank == 0) log_write_end();

    if (my_rank == 0) {
      time = (double)(end_t - start_t);
      flop = 2.0/3.0*(double)num_equations*num_equations*num_equations +
	2.0*(double)num_equations*num_equations;
      printf("Solve took %lld cycles for %d equations (= %.4f GFLOP/s)\n",
	     end_t - start_t, num_equations, flop/time);

      log_write_result_double("flop_count", flop);
      log_write_result_double("time", time);
    }

    if (my_rank == 0) {
      ret = verify(M, x, num_equations, random_seed);
      printf("verify returned %d\n", ret);
    } else {
      ret = 0;
    }

  }

  return ret;
}
