/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef DMA_LINPACK_H
#define DMA_LINPACK_H

#include <stddef.h>
#include <stdint.h>

/* Data layout constants. */

#define SUBBLOCK_SIZE 4
#define BLOCK_SIZE 16
#define PANEL_SIZE (8*BLOCK_SIZE)
#define MATRIX_SIZE (63*PANEL_SIZE)

#define M_L2_ROWS 64
#define M_L2_COLS (PANEL_SIZE)

#define L_L1_ROWS 8
#define U_L1_COLS 4


/* Index functions. */

/* M is blocked into  BLOCK_SIZE*BLOCK_SIZE size blocks. */
static inline size_t
index_m(const size_t row, const size_t col)
{
  return  ( (row/BLOCK_SIZE)*BLOCK_SIZE*MATRIX_SIZE + 
	    (col/BLOCK_SIZE)*BLOCK_SIZE*BLOCK_SIZE +
	    (row%BLOCK_SIZE)*BLOCK_SIZE +	
	    col%BLOCK_SIZE );
}

/* We use the leftmost columns of M to store L. */
/* Since we always work on whole panels, we use PANEL_SIZE^2
   chunks of M to store L.
   Within these blocks, the matrix is stored transposed in 
   BLOCK_SIZE^2 blocks. */
static inline size_t
index_l(const size_t row, const size_t col)
{
  return ( (row/PANEL_SIZE)*PANEL_SIZE*MATRIX_SIZE +
	   (col/PANEL_SIZE)*BLOCK_SIZE*PANEL_SIZE +
	   (col%PANEL_SIZE)/BLOCK_SIZE*BLOCK_SIZE*MATRIX_SIZE +
	   (row%PANEL_SIZE)/BLOCK_SIZE*BLOCK_SIZE*BLOCK_SIZE +
	   (col%BLOCK_SIZE)*BLOCK_SIZE +
	   (row%BLOCK_SIZE) );
}

static inline size_t
index_l_l3(const size_t row, const size_t col)
{
  return ( (row/BLOCK_SIZE)*BLOCK_SIZE*PANEL_SIZE +
	   col*BLOCK_SIZE +
	   row%BLOCK_SIZE );
}

/* M_l2 is blocked into  BLOCK_SIZE*BLOCK_SIZE size blocks. */
static inline size_t
index_m_l2(const size_t row, const size_t col, const size_t width)
{
  return ( (row/BLOCK_SIZE)*BLOCK_SIZE*width +
	   (col/BLOCK_SIZE)*BLOCK_SIZE*BLOCK_SIZE + 
	   (row%BLOCK_SIZE)*BLOCK_SIZE +
	   col%BLOCK_SIZE );
}

/* L_l2 is blocked into SUBBLOCK_SIZE panels. */
static inline size_t
index_l_l2(const size_t row, const size_t col)
{
  return ( (row/SUBBLOCK_SIZE)*SUBBLOCK_SIZE*PANEL_SIZE +
	   col*SUBBLOCK_SIZE +
	   row%SUBBLOCK_SIZE );
}

/* U_l2 is blocked into SUBBLOCK_SIZE panels. */
static inline size_t
index_u_l2(const size_t row, const size_t col)
{
  return ( (col/SUBBLOCK_SIZE)*SUBBLOCK_SIZE*PANEL_SIZE +
	   row*SUBBLOCK_SIZE +
	   col%SUBBLOCK_SIZE );
}

/* Normal functions. */

void init_problem(double * restrict m, size_t num_equations, uint64_t seed);

int verify(double * restrict m,
	   const double * restrict x,
	   size_t num_equations, 
	   uint64_t seed);

void back_substitute(double * restrict x,
		     const double * restrict m,
		     size_t num_equations);

void solve(double * restrict x,
	   double * restrict m,
	   size_t num_equations);

void factorize(double * restrict m, size_t num_equations);

void factorize_panel(size_t * restrict pivot_rows,
		     double * restrict pivot_scales,
		     double * restrict m,
		     size_t rows,
		     size_t panel_width);

void fill_u_panel(double * restrict m,
		  const size_t * restrict pivot_rows,
		  size_t rows,
		  size_t cols);

void fill_u_panel_l2(double * restrict u_l2,
		     double * restrict m,
		     const size_t * restrict pivot_rows,
		     size_t rows,
		     size_t cols);

void fill_l_head_l2(double * restrict l_head,
		    const double * restrict m,
		    size_t rows,
		    size_t cols);

void save_u_panel_l2(double * restrict m,
		     const double * restrict u_l2,
		     size_t rows,
		     size_t cols);

void update_u_panel(double * restrict m, size_t panel_width, size_t cols);

void update_u_panel_l2(double * restrict u_l2,
		       const double * restrict l_head,
		       size_t panel_width,
		       size_t cols);

uint64_t update_m_panel(double * restrict m, 
			const size_t * restrict pivot_rows,
			size_t panel_width, 
			size_t cols);
void setup_m_panel();

void block_update(double *mnew, 
		  const double *mold,
                  const double *l, 
		  const double *u,
                  unsigned block_count,
		  unsigned panel_count);

void update_block_l2(double *mnew,
		     double *m_l1_calc,
		     double *m_l1_load,
                     double *l_l1,
                     const double *l_l2_next,
		     double *u_l1,
		     const double *m_l2_load,
		     const double *l_l2_load,
		     const double *u_l2,
		     unsigned panel_size,
		     int u_last_adjust);

#if 0
void update_block_l2(double *mnew, 
		     const double *mold,
		     const double *l, 
		     const double *u,
		     unsigned block_count,
		     unsigned panel_count);
#endif

unsigned update_block_idma(double * restrict m_l2_new,
			   const double * restrict m_l2_old,
			   const double * restrict u_l2,
			   const double * restrict l_l2,
			   double * restrict m_l1,
			   double * restrict u_l1,
			   double * restrict l_l1,
			   unsigned block_rows,
			   unsigned block_cols);

unsigned update_block_idma_new(double * restrict m_l2,
			       const double * restrict u_l2,
			       const double * restrict l_l2,
			       double * restrict m_l1,
			       double * restrict u_l1,
			       double * restrict l_l1,
			       unsigned cnt_reload);

unsigned update_block_idma_row(double * restrict m_l2,
			       const double * restrict u_l2,
			       const double * restrict l_l2,
			       double * restrict m_l1,
			       double * restrict u_l1,
			       double * restrict l_l1,
			       unsigned cnt_reload);

void update_block_setup_edma(void * m_l2, void * l_l2, void * l_l3);

void update_block_edma(double * restrict m,
		       double * restrict l,
		       double * restrict l_l3,
		       double * restrict m_l2,
		       double * restrict u_l2,
		       double * restrict l_l2,
		       double * restrict m_l1,
		       double * restrict u_l1,
		       double * restrict l_l1,
		       size_t rows,
		       size_t cols);

void init_l_xfer(double * restrict m, unsigned row, unsigned col);
unsigned wait_l_xfer();
void transpose_l_panel(double * restrict m, size_t panel_width, size_t rows);
void setup_panel(unsigned core, unsigned width);

#endif
