/**
   Test L-panel EDMA transfer to L2 SRAM for DMA based linpack.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <stdio.h>
#include <stdlib.h>

#include <cache.h>
#include <barrier.h>
#include <tsc.h>
#include <log.h>

/* Various matrices needed, static allocation! */

/* The big data matrix [A||B]. */
#pragma DATA_SECTION(M, "matrix")
double M[MATRIX_SIZE*MATRIX_SIZE];

/* The solution vector. */
#pragma DATA_SECTION(x, "matrix")
double x[MATRIX_SIZE];

/* The shared (MCSM) buffers. */
extern double L_l3[PANEL_SIZE*PANEL_SIZE/2];

/* The L2 cache buffers. */
extern double M_l2[2*M_L2_ROWS*M_L2_COLS];
extern double U_l2[2*PANEL_SIZE*M_L2_COLS];
extern double L_l2[2*M_L2_ROWS*PANEL_SIZE];

/* The L1 cache buffers. */
extern double L_l1[2*L_L1_ROWS*PANEL_SIZE];
extern double U_l1[2*PANEL_SIZE*U_L1_COLS];
extern double M_l1[2*BLOCK_SIZE*BLOCK_SIZE];

/******************************************************************************/

static inline double
init_m(size_t row, size_t col)
{
  return ((double)(row)*10000.0 + (double)(col));
}

int
main(int argc, char **argv)
{
  size_t row;
  size_t col;
  size_t block_row;

  unsigned my_rank;
  unsigned num_ranks;
  unsigned core;

  double actual, desired;

  my_rank = cache_get_mm_id();
  num_ranks = cache_get_mm_num();

  if (my_rank != 0) return 0;

  cache_l1d_set_size(4);	/* 32kB L1D cache, 0kB SRAM*/
  cache_l2_set_size(0);		/* 0kB L2, 512kB SRAM */

  /* Initialize DDR3 matrix. */
  for(row=0; row<MATRIX_SIZE; ++row) 
    for(col=0; col<MATRIX_SIZE; ++col)
      M[index_m(row, col)] = init_m(row, col);

  /* Setup the EDMA transfers stuff. */
  update_block_setup_edma(M_l2, L_l2, L_l3);
  for(core=0; core<8;++core) 
    setup_panel(core, 8);

#if 0
  transpose_l_panel(M + index_m(PANEL_SIZE, 0),
		    PANEL_SIZE,
		    MATRIX_SIZE - PANEL_SIZE);
#endif

  for(block_row = PANEL_SIZE; block_row<MATRIX_SIZE; block_row+=M_L2_ROWS) {

    /* Init the L3 and L2 buffers. */
    for(row=0; row<2*M_L2_ROWS*PANEL_SIZE; ++row) 
      L_l3[row] = -3.0;
    
    for(row=0; row<2*M_L2_ROWS*PANEL_SIZE; ++row)
      L_l2[row] = -2.0;

    /* Do the DMA transfer. */
    init_l_xfer(M, block_row, 0);
    wait_l_xfer();

    /* Check the result. */
    for(row=0; row<M_L2_ROWS; ++row) {
      for(col=0; col<PANEL_SIZE; ++col) {
	desired = init_m(block_row + row, col);
	actual = L_l2[index_l_l2(row + block_row%PANEL_SIZE, col)];
	if (fabs(desired - actual) > 1E-10) {
	  printf("OUTCH, row %d col %d wanted %f got %f !\n",
		 block_row + row, col, desired, actual);
	}
      }
    }
  }

  printf("All done!\n");

  return 0;
}
