/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <stdio.h>
#include <stdlib.h>

#include <cache.h>
#include <barrier.h>
#include <idma.h>
#include <tsc.h>

/* Various matrices needed, static allocation! */

/* The big data matrix [A||B]. */
#pragma DATA_SECTION(M, "matrix")
double M[MATRIX_SIZE*MATRIX_SIZE];

/* The solution vector. */
#pragma DATA_SECTION(x, "matrix")
double x[MATRIX_SIZE];

/* The L2 cache buffers. */
#pragma DATA_SECTION(M_l2, "l2")
double M_l2[2*M_L2_ROWS*M_L2_COLS];
#pragma DATA_SECTION(U_l2, "l2")
double U_l2[2*PANEL_SIZE*M_L2_COLS];
#pragma DATA_SECTION(L_l2, "l2")
double L_l2[2*M_L2_ROWS*PANEL_SIZE];

/* The L1 cache buffers. */
#pragma DATA_SECTION(L_l1, "l1d")
double L_l1[2*L_L1_ROWS*PANEL_SIZE];
#pragma DATA_SECTION(U_l1, "l1d")
double U_l1[2*PANEL_SIZE*U_L1_COLS];
#pragma DATA_SECTION(M_l1, "l1d")
double M_l1[2*BLOCK_SIZE*BLOCK_SIZE];

static inline double
init_m(unsigned row, unsigned col)
{
  return ((double)row + 1.0)*(col + 1.0);
  /* return 0.0; */
}

static inline double
init_u(unsigned row, unsigned col)
{
  return ((double)row + 1.0)*(col + 1.0);
  /*return 1.0;*/
}

static inline double
init_l(unsigned row, unsigned col)
{
  return ((double)row + 3.0)*(col + 3.0);
  /* return 1.0; */
}

/******************************************************************************/

int
main(int argc, char **argv)
{
  size_t num_equations = 9*PANEL_SIZE - 1;
  uint64_t random_seed = 1234;
  size_t i;
  int ret;
  unsigned my_rank;
  size_t row, col;
  double actual, desired;
  unsigned wait_cnt;
  unsigned errors;
  uint64_t end_t, start_t;
  size_t block_cols;
  size_t width;

  my_rank = cache_get_mm_id();
  wait_cnt = 0;

  cache_l1d_set_size(1);	/* 4kB L1D cache, 28kB SRAM */
  cache_l2_set_size(0);		/* 0kB L2, 512kB SRAM */ 

  if (my_rank == 0) {
    for(block_cols=1; block_cols<9; ++block_cols) {
      width = block_cols*BLOCK_SIZE;

      for(i=0; i<2*M_L2_ROWS*M_L2_COLS; ++i) M_l2[i]=0.0;
      for(i=0; i<2*PANEL_SIZE*M_L2_COLS; ++i) U_l2[i]=0.0;
      for(i=0; i<2*M_L2_ROWS*PANEL_SIZE; ++i) L_l2[i]=0.0;

      for(row=0; row<2*M_L2_ROWS; ++row)
	for(col=0; col<PANEL_SIZE; ++col) 
	  L_l2[index_l_l2(row, col)] = init_l(row, col);

      for(row=0; row<PANEL_SIZE; ++row)
	for(col=0; col<2*M_L2_COLS; ++col) 
	  U_l2[index_u_l2(row, col)] = init_u(row, col);

      for(row=0; row<2*M_L2_ROWS; ++row)
	for(col=0; col<width; ++col) 
	  M_l2[index_m_l2(row, col, width)] = init_m(row, col);

      /* Do the IDMA dance to init M_l1 and L_l1 */
      /* Read the l_l1 matrix. */
      idma1_setup(L_l1, 
		  L_l2, 
		  sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		  0, 
		  0, 
		  7);
    
      /* Read the m_l1 matrix. */
      idma1_setup(M_l1, 
		  M_l2, 
		  sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
		  0, 
		  0, 
		  7);

      /* Wait until one slot is free again. */
      while ((idma1_status() & 0x02) != 0) wait_cnt++;

      /* Bring in the first column panel for the U matrix. */
      idma1_setup(U_l1, 
		  U_l2, 
		  2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		  0, 
		  0, 
		  7);
      /* Wait until one slot is free again. */
      mfence();
      while ((idma1_status() & 0x02) != 0) wait_cnt++;

      /* Zero pad so that we can see */
      /*for(i=PANEL_SIZE*U_L1_COLS; i<2*PANEL_SIZE*U_L1_COLS; ++i) U_l1[i] = 0.0;*/
      for(i=PANEL_SIZE*L_L1_ROWS; i<2*PANEL_SIZE*L_L1_ROWS; ++i) L_l1[i] = 0.0;
      for(i=BLOCK_SIZE*BLOCK_SIZE; i<2*BLOCK_SIZE*BLOCK_SIZE; ++i) M_l1[i] = 0.0;

#if 0
#if 0
      start_t = tsc_get();
#endif 
      wait_cnt = update_block_idma_new(M_l2,
				       U_l2,
				       L_l2,
				       M_l1,
				       U_l1,
				       L_l1,
				       ( (3<<24) | 
					 ((block_cols - 1)<<16) | 
					 (15<<8) | 
					 PANEL_SIZE ));
#if 0
      end_t = tsc_get();
      printf("update_block_idma %llu cycles, wait_cnt %u\n", 
	     end_t - start_t, 
	     wait_cnt);
#endif

      /* Wait until one slot is free again. */
      while ((idma1_status() & 0x02) != 0) wait_cnt++;

      /* Do the IDMA dance to init M_l1 and L_l1 */
      /* Read the l_l1 matrix. */
      idma1_setup(L_l1, 
		  L_l2, 
		  2*sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		  0, 
		  0, 
		  7);

#if 0    
      /* Read the m_l1 matrix. */
      idma1_setup(M_l1, 
		  M_l2, 
		  2*sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
		  0, 
		  0, 
		  7);
#endif

      /* Wait until one slot is free again. */
      while ((idma1_status() & 0x02) != 0) wait_cnt++;

      /* Bring in the first column panel for the U matrix. */
      idma1_setup(U_l1, 
		  U_l2, 
		  2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		  0, 
		  0, 
		  7);
      /* Wait until one slot is free again. */
      mfence();
      while ((idma1_status() & 0x02) != 0) wait_cnt++;
#endif

      start_t = tsc_get();
      wait_cnt = update_block_idma_new(M_l2,
				       U_l2,
				       L_l2,
				       M_l1,
				       U_l1,
				       L_l1,
				       ( (3<<24) | 
					 ((block_cols - 1)<<16) | 
					 (15<<8) | 
					 PANEL_SIZE ) );
      end_t = tsc_get();

      printf("update_block_idma %llu cycles, width %u\n", 
	     end_t - start_t, 
	     width);

      /* Check that the result is good. */
      coherence_flush_all();
      errors = 0;
      for(row=0; row<M_L2_ROWS; ++row) {
	for(col=0; col<width; ++col) {
	  if (row<64) {
#if 0
	    if (((row/BLOCK_SIZE)*block_cols + col/BLOCK_SIZE)%2 == 0) {
	      desired = init_m(row % 16, col % 16);
	    } else {
	      if (width<32) desired = init_m((row % 16) + 16, col % 16);
	      else desired = init_m(row % 16, col % 16 + 16);
	    }
#endif
	    desired = init_m(row, col);

	    for(i=0; i<PANEL_SIZE; ++i) 
	      desired -= init_l(row, i)*init_u(i, col);
	  } else {
	    desired = init_m(row, col);
	  }

	  actual = M_l2[index_m_l2(row, col, width)]; 

	  if (fabs(actual-desired) > 1E-10) {
	    if (errors < 20) {
	      printf("OUTCH, row %3u col %3u wanted %10.6f got %10.6f\n",
		     row, col, desired, actual);
	    }
	    ++errors;
	  }
	}
      }

#if 0
      for(row=0; row<BLOCK_SIZE; ++row) {
	for(col=0; col<BLOCK_SIZE; ++col) {
	  desired = init_m(row, col + BLOCK_SIZE);
	  actual = M_l1[row*BLOCK_SIZE + col + BLOCK_SIZE*BLOCK_SIZE];
	  if (fabs(actual-desired) > 1E-10) {
	    printf("OUTCH M_l1, row %3u col %3u wanted %10.6f got %10.6f\n",
		   row, col, desired, actual);
	    ++errors;
	  }
	}
      }
#endif

      printf("Width %u found %u errors!\n", block_cols*BLOCK_SIZE, errors);
    }
  }
  barrier();

  return ret;
}
