/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"

#include <edma3.h>
#include <tsc.h>
#include <cache.h>
#include <barrier.h>
#include <idma.h>

/* The EDMA3 controller to use 1 = 0x02720000 */
#define CC 1

/* For the L panel update, use 0 = 0x02700000 */
#define LCC 0

static void setup_m_edma(void *l2_buffer);
static void setup_l_edma(void *l2_buffer, void *l3_buffer);

void
update_block_setup_edma(void * m_l2, void * l_l2, void * l_l3)
{
  setup_m_edma(m_l2);
  setup_l_edma(l_l2, l_l3);

  /* Step 3: Disable interrupts */
  edma3cc_registers[CC].global.iecr = 0xfffffffful;
  edma3cc_registers[CC].global.iecrh = 0xfffffffful;

  edma3cc_registers[LCC].global.iecr = 0xfffffffful;
  edma3cc_registers[LCC].global.iecrh = 0xfffffffful;
  
  
  /* Clear the event registers. */
  edma3cc_registers[CC].global.esr = 0;
  edma3cc_registers[CC].global.esrh = 0;
  edma3cc_registers[CC].global.ecr = 0xfffffffful;
  edma3cc_registers[CC].global.ecrh = 0xfffffffful;
  edma3cc_registers[CC].global.secr = 0xfffffffful;
  edma3cc_registers[CC].global.secrh = 0xfffffffful;
  edma3cc_registers[CC].global.icr = 0xfffffffful;
  edma3cc_registers[CC].global.icrh = 0xfffffffful;

  edma3cc_registers[LCC].global.esr = 0;
  edma3cc_registers[LCC].global.esrh = 0;
  edma3cc_registers[LCC].global.ecr = 0xfffffffful;
  edma3cc_registers[LCC].global.ecrh = 0xfffffffful;
  edma3cc_registers[LCC].global.secr = 0xfffffffful;
  edma3cc_registers[LCC].global.secrh = 0xfffffffful;
  edma3cc_registers[LCC].global.icr = 0xfffffffful;
  edma3cc_registers[LCC].global.icrh = 0xfffffffful;
}



/******************************************************************************/
/* M block transfer stuff                                                     */
/******************************************************************************/

/**
   Setup the EDMA CC for the transfer pattern.

   This involves 4*8 read and 4*8 write transfers for all 8 cores.

   The cores are divided into two disjunct sets that each open their own
   bank.

   Transfers are interleaved per core:

    0: L2->DDR Row 0 Core 0
    1: L2->DDR Row 0 Core 1
   ...
    7: L2->DDR Row 0 Core 7
    8: L2->DDR Row 1 Core 0
   ...
   31: L2->DDR Row 3 Core 7
   32: DDR->L2 Row 0 Core 0
   ...
   63: DDR->L2 Row 3 Core 7

   Odd  cores -> TC0 (banks 0, 2, 4, 6)
   Even cores -> TC1 (banks 1, 3, 5, 7)

   Transfers chain trigger to the next row in the same core and
   direction with wrap-aroud at the end. The last transfer in the 
   last row informs the application about the completion of the transfer.

   All transfers a syncronized for a single block size.

   The buffer needs to have 4 rows of 16 blocks! (4*8, A+B)
*/
void
setup_m_edma(void *l2_buffer)
{
  unsigned i;
  unsigned core;
  unsigned row;
  uint32_t block_size = 16*16*sizeof(double);
  uint32_t buffer_start;
  volatile struct edma3_param_set *param;

  /* Init the controller. */
  edma3_param_init(CC);
  for(i=0; i<8; ++i) edma3_set_qdma_channel(CC, i, 511, 0);
  for(i=0; i<64; ++i) edma3_set_dma_channel(CC, i, 511);

  /* Setup the channel and queue mapping. */
  for(row=0; row<4; ++row) {
    for(core=0; core<8; ++core) {
      edma3_set_dma_channel(CC, row*8 + core, row*8 + core);
      edma3_set_dma_queue(CC, row*8 + core, 2*(core%2));
      edma3_set_dma_channel(CC, row*8 + core + 32, row*8 + core + 32);
      edma3_set_dma_queue(CC, row*8 + core + 32, 2*(core%2));
    }
  }

  /* Setup the parameter set */
  for(row=0; row<4; ++row) {
    buffer_start = ((uint32_t)l2_buffer) + row*block_size*16; 
    for(core=0; core<8; ++core) {

      /* L2 -> DDR */
      param =  edma3cc_registers[CC].param + row*8 + core;

      param->opt = 
	((row==3)?edma3_opt_tcinten:edma3_opt_tcchen) | 
	edma3_opt_itcchen | 
	edma3_calc_opt_tcc(((row + 1)%4)*8 + core);
      param->src = (uint32_t)0;
      param->abcnt = edma3_calc_abcnt(2048, 8);
      param->dst = (uint32_t)0;
      param->bidx = edma3_calc_idx(2048, 8*2048);
      param->link_bcntrld = edma3_calc_link((row*8 + core)*32, 8);
      param->cidx = edma3_calc_idx(0, 0);
      param->ccnt = 1;

      /* DDR -> L2 */
      param =  edma3cc_registers[CC].param + row*8 + core + 32;
      param->opt = 
	((row==3)?edma3_opt_tcinten:edma3_opt_tcchen) | 
	edma3_opt_itcchen | 
	edma3_calc_opt_tcc(((row + 1)%4)*8 + core + 32);
      param->src = (uint32_t)0;
      param->abcnt = edma3_calc_abcnt(2048, 8);
      param->dst = (uint32_t)0;
      param->bidx = edma3_calc_idx(8*2048, 2048);
      param->link_bcntrld = edma3_calc_link((row*8 + core + 32)*32, 8);
      param->cidx = edma3_calc_idx(0, 0);
      param->ccnt = 1;
    }
  }
}

void
setup_panel(unsigned core, unsigned width)
{
  unsigned row;
  volatile struct edma3_param_set *param; 

  for(row=0; row<4; ++row) {
    /* L2->DDR */
    param =  edma3cc_registers[CC].param + row*8 + core;
    param->abcnt = edma3_calc_abcnt(2048, width);
    param->link_bcntrld = edma3_calc_link((row*8 + core)*32, width);

    /* DDR->L2*/
    param =  edma3cc_registers[CC].param + row*8 + core + 32;
    param->abcnt = edma3_calc_abcnt(2048, width);
    param->link_bcntrld = edma3_calc_link((row*8 + core + 32)*32, width);
  }
}

void setup_row(unsigned core,
	       unsigned width,
	       void *input_start, 
	       void *output_start, 
	       void *l2_buffer)
{
  unsigned row;
  const uint32_t block_size = BLOCK_SIZE*BLOCK_SIZE*sizeof(double);
  const uint32_t row_stride = MATRIX_SIZE*BLOCK_SIZE*sizeof(double);
  uint32_t buffer_start;
  volatile struct edma3_param_set *param;
 
  for(row=0; row<4; ++row) {
    buffer_start = 
      ((uint32_t)l2_buffer) + row*block_size*width | 
      0x10000000ul | 0x01000000ul*core;
    
    /* L2->DDR */
    param = edma3cc_registers[CC].param + row*8 + core;
    param->src = buffer_start;
    param->dst = ((uint32_t)output_start) + row*row_stride;
    param->abcnt = edma3_calc_abcnt(2048, width);

    /* DDR->L2 */
    param = edma3cc_registers[CC].param + row*8 + core + 32;
    param->src = ((uint32_t)input_start) + row*row_stride;
    param->dst = buffer_start;
    param->abcnt = edma3_calc_abcnt(2048, width);
  }
}

static inline void 
edma_read()
{
  const uint32_t event=0xff;
  unsigned wait_cnt=0;
  
 /* Kick off the transfer. */
  edma3cc_registers[CC].global.esrh = event;
}

static inline void
edma_read_core(unsigned core)
{
  /* Kick off the transfer. */
  edma3cc_registers[CC].global.esrh = 1u << core;
}

static inline unsigned
edma_wait_read()
{
  const uint32_t event=0xff;
  unsigned wait_cnt=0;

  /* Wait until done. */
  while ((edma3cc_registers[CC].global.iprh & event) != event) wait_cnt++;
  
  /* Clear the interrupts. */
  edma3cc_registers[CC].global.icrh = event; 
  
  return wait_cnt;
}

static inline unsigned
edma_wait_read_core(unsigned core)
{
  uint32_t event = 1 << core;
  unsigned wait_cnt = 0;

  /* Wait until done. */
  while ((edma3cc_registers[CC].global.iprh & event) != event) wait_cnt++;
  
  /* Clear the interrupts. */
  edma3cc_registers[CC].global.icrh = event; 
  
  return wait_cnt; 
}

static inline void 
edma_write()
{
  const uint32_t event=0xff;
  unsigned wait_cnt=0;
  
  /* Kick off the transfer. */
  edma3cc_registers[CC].global.esr = event;
}

static inline void
edma_write_core(unsigned core)
{
  /* Kick off the transfer. */
  edma3cc_registers[CC].global.esr = 1u << core;
}

static inline unsigned
edma_wait_write()
{
  const uint32_t event=0xff;
  unsigned wait_cnt=0;

  /* Wait until done. */
  while ((edma3cc_registers[CC].global.ipr & event) != event) wait_cnt++;
  
  /* Clear the interrupts. */
  edma3cc_registers[CC].global.icr = event; 

  return wait_cnt;
}

static inline unsigned
edma_wait_write_core(unsigned core)
{
  uint32_t event = 1 << core;
  unsigned wait_cnt = 0;

  /* Wait until done. */
  while ((edma3cc_registers[CC].global.ipr & event) != event) wait_cnt++;
  
  /* Clear the interrupts. */
  edma3cc_registers[CC].global.icr = event; 
  
  return wait_cnt; 
}

static inline unsigned
edma_wait_write_all_done()
{
  unsigned wait_cnt;
  
  wait_cnt = 0;
  while (edma3cc_registers[CC].global.ipr != 0) wait_cnt++;

  return wait_cnt;
}

/******************************************************************************/
/* L-panel EDMA update                                                        */
/******************************************************************************/


/*
  Setup the EDMA channel controller/PaRAM for the L panel update.

  The L panel is loaded in two steps: DDR->MSMC->L2(x8)

  The first step rearranges the data from the 128x128 block that is
  block in 16x16 size subblocks to the MSMC buffer which is 128x64
  doubles big and arraned in 4 16 row long panels.

  The second step rearranges the data from the 128x64 MSMC into the
  128x64 L2 buffer. Here the data layout is changed to 4 long panels
  instead.

  PaRAM entries:

  000 ... 007  DDR -> MCSM, row%128 == 0
  008 ... 039  MCSM -> L2, row%128 == 0

  064 ... 071  DDR -> MCSM, row%128 == 64
  072 ... 103  MCSM -> L2, row%128 == 0

  112          Normal Channel 0
  127          Null Transfer, all unused events and QDMA channels.
 */
void
setup_l_edma(void *l2_buffer, void *l3_buffer)
{
  unsigned i;
  unsigned col, row;
  unsigned core;
  unsigned ofs;
  int last_col;
  volatile struct edma3_param_set *param;
  
  /* Init the controller. */
  edma3_param_init(LCC);
  for(i=0; i<8; ++i) edma3_set_qdma_channel(LCC, i, 127, 0);
  for(i=0; i<16; ++i) edma3_set_dma_channel(LCC, i, 127);
  
  /* Setup channel and queue mapping. */
  edma3_set_dma_channel(LCC, 0, 112);
  edma3_set_dma_queue(LCC, 0, 0);

  /* Initialize the parameter sets. */
#if 0

  for(col=0; col<PANEL_SIZE/BLOCK_SIZE; col++) {
    /* row 0 */
    param = edma3cc_registers[LCC].param + col;
    
    param->opt =
      edma3_opt_tcchen |
      edma3_opt_tccmode |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);
    param->src = 0;
    param->abcnt = edma3_calc_abcnt(BLOCK_SIZE*BLOCK_SIZE*sizeof(double),
				    PANEL_SIZE/(2*BLOCK_SIZE));
    param->dst = ( ((uint32_t)l3_buffer) + 
		   col*BLOCK_SIZE*BLOCK_SIZE*sizeof(double) );
    param->bidx = edma3_calc_idx(BLOCK_SIZE*BLOCK_SIZE*sizeof(double),
				 BLOCK_SIZE*PANEL_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((col + 1)*32, 0);
    param->cidx = edma3_calc_idx(0, 0);
    param->ccnt = 1;

    /* row 64 */
    param = edma3cc_registers[LCC].param + col + 64;
    
    param->opt =
      edma3_opt_tcchen |
      edma3_opt_tccmode |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);
    param->src = 0;
    param->abcnt = edma3_calc_abcnt(BLOCK_SIZE*BLOCK_SIZE*sizeof(double),
				    PANEL_SIZE/(2*BLOCK_SIZE));
    param->dst = ( ((uint32_t)l3_buffer) + 
		   col*BLOCK_SIZE*BLOCK_SIZE*sizeof(double) );
    param->bidx = edma3_calc_idx(BLOCK_SIZE*BLOCK_SIZE*sizeof(double),
				 BLOCK_SIZE*PANEL_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((col + 65)*32, 0);
    param->cidx = edma3_calc_idx(0, 0);
    param->ccnt = 1;
  }

  for(row=0; row<M_L2_ROWS/BLOCK_SIZE; row++) {
    last_col = row==(M_L2_ROWS/BLOCK_SIZE - 1);

    for(core=0; core<8; core++) {
      /* row 0 */
      param = edma3cc_registers[LCC].param + row*8 + core + 8;

      param->opt = 
	edma3_opt_syncdim |
	((last_col && core==7)?edma3_opt_tcinten:edma3_opt_tcchen) |
	edma3_opt_itcchen |
	((last_col && core==7)?0:edma3_opt_tccmode) |
	edma3_calc_opt_tcc(0);
      param->src = ((uint32_t)l3_buffer) + 
	index_l_l3(row*BLOCK_SIZE, 0)*sizeof(double);
      param->abcnt = edma3_calc_abcnt(4*sizeof(double), BLOCK_SIZE/4);
      param->dst = ((uint32_t)l2_buffer) +
	index_l_l2(row*BLOCK_SIZE, 0)*sizeof(double) |
	0x10000000ul | 0x01000000ul*core;
      param->bidx = 
	edma3_calc_idx(4*sizeof(double), 4*PANEL_SIZE*sizeof(double));
      param->link_bcntrld = 
	edma3_calc_link((last_col && core==7)?64*32:(row*8 + core + 9)*32, 0);
      param->cidx = edma3_calc_idx(BLOCK_SIZE*sizeof(double), 4*sizeof(double));
      param->ccnt = PANEL_SIZE;

      /* row 64 */
      param = edma3cc_registers[LCC].param + row*8 + core + 72;

      param->opt = 
	edma3_opt_syncdim |
	((last_col && core==7)?edma3_opt_tcinten:edma3_opt_tcchen) |
	edma3_opt_itcchen |
	((last_col && core==7)?0:edma3_opt_tccmode) |
	edma3_calc_opt_tcc(0);
      param->src = ((uint32_t)l3_buffer) + 
	index_l_l3(row*BLOCK_SIZE, 0)*sizeof(double);
      param->abcnt = edma3_calc_abcnt(4*sizeof(double), BLOCK_SIZE/4);
      param->dst = ((uint32_t)l2_buffer) +
	index_l_l2(row*BLOCK_SIZE + M_L2_ROWS, 0)*sizeof(double) |
	0x10000000ul | 0x01000000ul*core;
      param->bidx = 
	edma3_calc_idx(4*sizeof(double), 4*PANEL_SIZE*sizeof(double));
      param->link_bcntrld = 
	edma3_calc_link((last_col && core==7)?0:(row*8 + core + 73)*32, 0);
      param->cidx = edma3_calc_idx(BLOCK_SIZE*sizeof(double), 4*sizeof(double));
      param->ccnt = PANEL_SIZE;
    }
  }
#else
  
  /* Step 1: DDR -> L3 in "normal" C layout, all except last early. */
  for(row=0; row<M_L2_ROWS/BLOCK_SIZE; ++row) {
    /* row  0 : ofs = 0 */
    param = edma3cc_registers[LCC].param + row;
    
    param->opt =
      edma3_opt_tcchen |
      edma3_opt_itcchen |
      (row<(M_L2_ROWS/BLOCK_SIZE - 1)?edma3_opt_tccmode:0) |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);

    param->src = 0;
    param->abcnt = 
      edma3_calc_abcnt(BLOCK_SIZE*sizeof(double), BLOCK_SIZE);
    param->dst = 
      ( ((uint32_t)l3_buffer) + row*PANEL_SIZE*BLOCK_SIZE*sizeof(double) );
    param->bidx = edma3_calc_idx(BLOCK_SIZE*sizeof(double),
				 PANEL_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((row + 1)*32, BLOCK_SIZE);
    param->cidx = 
      edma3_calc_idx(BLOCK_SIZE*BLOCK_SIZE*sizeof(double), 
		     BLOCK_SIZE*sizeof(double));
    param->ccnt = PANEL_SIZE/BLOCK_SIZE;
    

    /* row 64 : ofs = 64 */
    param = edma3cc_registers[LCC].param + row + 64;
    
    param->opt =
      edma3_opt_tcchen |
      edma3_opt_itcchen |
      (row<(M_L2_ROWS/BLOCK_SIZE - 1)?edma3_opt_tccmode:0) |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);

    param->src = 0;
    param->abcnt = 
      edma3_calc_abcnt(BLOCK_SIZE*sizeof(double), BLOCK_SIZE);
    param->dst = 
      ( ((uint32_t)l3_buffer) + row*PANEL_SIZE*BLOCK_SIZE*sizeof(double) );
    param->bidx = edma3_calc_idx(BLOCK_SIZE*sizeof(double),
				 PANEL_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((row + 65)*32, BLOCK_SIZE);
    param->cidx = 
      edma3_calc_idx(BLOCK_SIZE*BLOCK_SIZE*sizeof(double), 
		     BLOCK_SIZE*sizeof(double));
    param->ccnt = PANEL_SIZE/BLOCK_SIZE;
  }

  /* Step 2: L3 -> L3, rearrange & transpose */
  for(row=0; row<SUBBLOCK_SIZE; ++row) {
    /* row 0 : ofs = 4 */
    param = edma3cc_registers[LCC].param + row + 4;

    param->opt =
      edma3_opt_tcchen |
      edma3_opt_itcchen |
      (row<(SUBBLOCK_SIZE - 1)?edma3_opt_tccmode:0) |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);

    param->src = 
      ((uint32_t)l3_buffer) + row*PANEL_SIZE*sizeof(double);
    param->abcnt = 
      edma3_calc_abcnt(sizeof(double), PANEL_SIZE);
    param->dst = 
      ( ((uint32_t)l3_buffer) + 
	M_L2_ROWS*PANEL_SIZE*sizeof(double) +
	row*sizeof(double));
    param->bidx = edma3_calc_idx(sizeof(double),
				 SUBBLOCK_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((row + 5)*32, PANEL_SIZE);
    param->cidx = 
      edma3_calc_idx(SUBBLOCK_SIZE*PANEL_SIZE*sizeof(double), 
		     SUBBLOCK_SIZE*PANEL_SIZE*sizeof(double));
    param->ccnt = M_L2_ROWS/SUBBLOCK_SIZE;

    /* row 64 : ofs = 68 */
    param = edma3cc_registers[LCC].param + row + 68;

    param->opt =
      edma3_opt_tcchen |
      edma3_opt_itcchen |
      (row<(SUBBLOCK_SIZE - 1)?edma3_opt_tccmode:0) |
      edma3_opt_syncdim |
      edma3_calc_opt_tcc(0);

    param->src = 
      ((uint32_t)l3_buffer) + row*PANEL_SIZE*sizeof(double);
    param->abcnt = 
      edma3_calc_abcnt(sizeof(double), PANEL_SIZE);
    param->dst = 
      ( ((uint32_t)l3_buffer) + 
	M_L2_ROWS*PANEL_SIZE*sizeof(double) +
	row*sizeof(double));
    param->bidx = edma3_calc_idx(sizeof(double),
				 SUBBLOCK_SIZE*sizeof(double));
    param->link_bcntrld = edma3_calc_link((row + 69)*32, PANEL_SIZE);
    param->cidx = 
      edma3_calc_idx(SUBBLOCK_SIZE*PANEL_SIZE*sizeof(double), 
		     SUBBLOCK_SIZE*PANEL_SIZE*sizeof(double));
    param->ccnt = M_L2_ROWS/SUBBLOCK_SIZE;

  }

  /* Step 3: L3 -> 8x L2 */
  for(core=0; core<8; core++) {
    /* row 0 : ofs = 8 */
    param = edma3cc_registers[LCC].param + core + 8;

    param->opt = 
      edma3_opt_syncdim |
	((core==7)?edma3_opt_tcinten:edma3_opt_tcchen) |
	edma3_opt_itcchen |
	((core==7)?0:edma3_opt_tccmode) |
	edma3_calc_opt_tcc(0);
    param->src = (uint32_t)l3_buffer + M_L2_ROWS*PANEL_SIZE*sizeof(double);
    param->abcnt = edma3_calc_abcnt(PANEL_SIZE*sizeof(double), M_L2_ROWS);
    param->dst = ((uint32_t)l2_buffer + 0x10000000ul) | 0x01000000ul*core;
    param->bidx = 
      edma3_calc_idx(PANEL_SIZE*sizeof(double), 
		     PANEL_SIZE*sizeof(double));
    param->link_bcntrld = 
      edma3_calc_link((core==7)?64*32:(core + 9)*32, 0);
    param->cidx = edma3_calc_idx(0, 0);
    param->ccnt = 1;
    
    /* row 64 : ofs = 72 */
    param = edma3cc_registers[LCC].param + core + 72;

    param->opt = 
      edma3_opt_syncdim |
      ((core==7)?edma3_opt_tcinten:edma3_opt_tcchen) |
      edma3_opt_itcchen |
      ((core==7)?0:edma3_opt_tccmode) |
      edma3_calc_opt_tcc(0);
    param->src = (uint32_t)l3_buffer + M_L2_ROWS*PANEL_SIZE*sizeof(double);
    param->abcnt = edma3_calc_abcnt(PANEL_SIZE*sizeof(double), M_L2_ROWS);
    param->dst = 
      ((uint32_t)l2_buffer + 
       PANEL_SIZE*M_L2_ROWS*sizeof(double) +
       0x10000000ul) | 0x01000000ul*core;
    param->bidx = 
      edma3_calc_idx(PANEL_SIZE*sizeof(double), 
		     PANEL_SIZE*sizeof(double));
    param->link_bcntrld = 
      edma3_calc_link((core==7)?0*32:(core + 73)*32, 0);
    param->cidx = edma3_calc_idx(0, 0);
    param->ccnt = 1;
  }

#endif

  edma3_param_copy(LCC, 112, 0);
}


void
init_l_xfer(double * restrict m, unsigned row, unsigned col)
{
  unsigned ofs, i;
  volatile struct edma3_param_set *param;

  ofs = (row%PANEL_SIZE)<M_L2_ROWS?0:64;

#if 0
  param = edma3cc_registers[LCC].param + 112;
  param->src = (uint32_t)(m + index_l(row, col));

  for(i=1; i<(PANEL_SIZE/BLOCK_SIZE); ++i) {
    param = edma3cc_registers[LCC].param + i + ofs;
    param->src = (uint32_t)(m + index_l(row, col + BLOCK_SIZE*i));
  }
#else
  param = edma3cc_registers[LCC].param + 112;
  param->src = (uint32_t)(m + index_m(row, col));

  for(i=1; i<(M_L2_ROWS/BLOCK_SIZE); ++i) {
    param = edma3cc_registers[LCC].param + i + ofs;
    param->src = (uint32_t)(m + index_m(row + BLOCK_SIZE*i, col));
  }
#endif

  /* Kick off the transfer. */
  edma3cc_registers[LCC].global.esr = 0x1;
}

unsigned
wait_l_xfer()
{
  const uint32_t event=0x1;
  unsigned wait_cnt=0;

  /* Wait until done. */
  while ((edma3cc_registers[LCC].global.ipr & event) != event) wait_cnt++;
  
  /* Clear the interrupts. */
  edma3cc_registers[LCC].global.icr = event; 
  
  return wait_cnt;
}


/******************************************************************************/
/* block update stuff                                                         */
/******************************************************************************/

void
update_block_edma(double * restrict m,
		  double * restrict l,
		  double * restrict l_l3,
		  double * restrict m_l2,
		  double * restrict u_l2,
		  double * restrict l_l2,
		  double * restrict m_l1,
		  double * restrict u_l1,
		  double * restrict l_l1,
		  size_t rows,
		  size_t cols)
{
  double *l2_calc, *l2_dma, *tmp;
  unsigned my_rank;
  unsigned core;
  unsigned wait_cnt;
  unsigned row;
  unsigned m_offset;

  my_rank = cache_get_mm_id();

  /* Prepare the transfers for the row. */
  l2_dma = m_l2;
  l2_calc = m_l2 + M_L2_ROWS*PANEL_SIZE;
  
  if (my_rank == 0) {
    for(core=0; core<8;++core) 
      setup_panel(core, 8);
  }

  if (my_rank == 0) {
    /* Step 1: DRAM -> L2(A) */
    for(core=0; core<8;++core) 
      setup_row(core, 
		cols/PANEL_SIZE, 
		m + index_m(0, core*BLOCK_SIZE), 
		m + index_m(0, core*BLOCK_SIZE), 
		l2_dma);
    edma_read();
    init_l_xfer(l, 0, 0);

    edma_wait_read();
    wait_l_xfer();

    /* Step 2: DRAM -> L2(B), calc A(row 0) */
    for(core=0; core<8;++core) 
      setup_row(core, 
		cols/PANEL_SIZE, 
		m + index_m(64, core*BLOCK_SIZE), 
		m + index_m(0, core*BLOCK_SIZE), 
		l2_calc);
    edma_read();
    init_l_xfer(l, M_L2_ROWS, 0);
  }
  barrier();

#if 0
  wait_cnt +=
    update_block_idma(l2_dma, l2_dma, 
		      u_l2, l_l2, 
		      m_l1, u_l1, l_l1, 
		      M_L2_ROWS, cols/8);
#else

#if 0
  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;

  /* Read the l_l1 matrix. */
  idma1_setup(l_l1, 
	      l_l2, 
	      sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
	      0, 
	      0, 
	      7);
    
  /* Read the m_l1 matrix. */
  idma1_setup(m_l1, 
	      l2_dma, 
	      sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
	      0, 
	      0, 
	      7);

  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;

  /* Bring in the first column panel for the U matrix. */
  idma1_setup(u_l1, 
	      u_l2, 
	      2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
	      0, 
	      0, 
	      7);

  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;
  
  wait_cnt +=
    update_block_idma_new(l2_dma, u_l2, l_l2, m_l1, u_l1, l_l1, 
			  ( ((M_L2_ROWS/BLOCK_SIZE - 1) << 24) |
			    ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			    (15 << 8) |
			    PANEL_SIZE ));
#else
  if (cols > 8*BLOCK_SIZE) {
    m_offset = BLOCK_SIZE*BLOCK_SIZE;
  } else {
    m_offset = BLOCK_SIZE*BLOCK_SIZE/2;
  }

  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;
  
  /* Read the l_l1 matrix. */
  idma1_setup(l_l1, 
	      l_l2, 
	      sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
	      0, 
	      0, 
	      7);
    
  /* Read the m_l1 matrix. */
  idma1_setup(m_l1, 
	      l2_dma, 
	      sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
	      0, 
	      0, 
	      7);

  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;
  
  /* Bring in the first column panel for the U matrix. */
  idma1_setup(u_l1, 
	      u_l2, 
	      2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
	      0, 
	      0, 
	      7);

  /* Read the m_l1 matrix. */
  idma1_setup(m_l1 + BLOCK_SIZE*BLOCK_SIZE/2, 
	      l2_dma + m_offset, 
	      sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
	      0, 
	      0, 
	      7);

  /* Wait until slots are free again. */
  mfence();
  while ((idma1_status() & 0x02) != 0) wait_cnt++;
  
  wait_cnt +=
    update_block_idma_row(l2_dma, u_l2, l_l2, m_l1, u_l1, l_l1, 
			  ( ((2*M_L2_ROWS/BLOCK_SIZE - 1) << 24) |
			    ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			    (7 << 8) |
			    PANEL_SIZE ));
  
#endif

#endif

#if 0
  barrier();
#endif

  if (my_rank == 0) {
    edma_wait_read();
    wait_l_xfer();
  }
  barrier();

  for(row=64; row<(rows - 64); row+=64) {

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1, 
		l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc, 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
		0, 
		0, 
		7);
    
    if (my_rank == 0) {
      for(core=0; core<8;++core) 
	setup_row(core, 
		  cols/PANEL_SIZE,
		  m + index_m(row + 64, core*BLOCK_SIZE), 
		  m + index_m(row - 64, core*BLOCK_SIZE), 
		  l2_dma);
      edma_write();		/* Write back old row */
    }

#if 0
    wait_cnt +=
      update_block_idma(l2_calc, l2_calc, 
			u_l2, 
			l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
			m_l1, u_l1, l_l1, 
			M_L2_ROWS/2, cols/8);
#else

#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1, 
		l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc, 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
		0, 
		0, 
		7);

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;

    /* Bring in the first column panel for the U matrix. */
    idma1_setup(u_l1, 
		u_l2, 
		2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		0, 
		0, 
		7);

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    wait_cnt +=
      update_block_idma_new(l2_calc, 
			    u_l2, 
			    l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
			    m_l1, 
			    u_l1, 
			    l_l1, 
			    ( ((M_L2_ROWS/(2*BLOCK_SIZE) - 1) << 24) |
			      ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			      (15 << 8) |
			      PANEL_SIZE ));
#else

#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1, 
		l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc, 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
		0, 
		0, 
		7);
#endif

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;

    /* Bring in the first column panel for the U matrix. */
    idma1_setup(u_l1 + PANEL_SIZE*U_L1_COLS, 
		u_l2 + PANEL_SIZE*U_L1_COLS, 
		sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		0, 
		0, 
		7);

    /* Read the m_l1 matrix. */
    idma1_setup(m_l1 + BLOCK_SIZE*BLOCK_SIZE/2, 
		l2_calc + m_offset, 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/4, 
		0, 
		0, 
		7);

#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
#endif
    
    wait_cnt +=
      update_block_idma_row(l2_calc, 
			    u_l2, 
			    l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
			    m_l1, 
			    u_l1, 
			    l_l1, 
			    ( ((M_L2_ROWS/BLOCK_SIZE - 1) << 24) |
			      ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			      (7 << 8) |
			      PANEL_SIZE ));

#endif

#endif

#if 0
    barrier();
#endif

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1,
		( l_l2 + 
		  (row%(2*M_L2_ROWS))*PANEL_SIZE +
		  (M_L2_ROWS)/2*PANEL_SIZE ), 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc + M_L2_ROWS/2*(cols/8), 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
		0, 
		0, 
		7);

    if (my_rank == 0) {
      edma_wait_write();

      edma_read();		/* Read in new row */
      init_l_xfer(l, row + M_L2_ROWS, 0);
    }

#if 0
    barrier();
#endif

#if 0
    wait_cnt +=
      update_block_idma(l2_calc + M_L2_ROWS/2*(cols/8), 
			l2_calc + M_L2_ROWS/2*(cols/8), 
			u_l2, 
			( l_l2 + 
			  (row%(2*M_L2_ROWS))*PANEL_SIZE +
			  (M_L2_ROWS)/2*PANEL_SIZE ), 
			m_l1, u_l1, l_l1, 
			M_L2_ROWS/2, cols/8);
#else

#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1,
		( l_l2 + 
		  (row%(2*M_L2_ROWS))*PANEL_SIZE +
		  (M_L2_ROWS)/2*PANEL_SIZE ), 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc + M_L2_ROWS/2*(cols/8), 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
		0, 
		0, 
		7);

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;

    /* Bring in the first column panel for the U matrix. */
    idma1_setup(u_l1, 
		u_l2, 
		2*sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		0, 
		0, 
		7);

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    wait_cnt +=
      update_block_idma_new(l2_calc + M_L2_ROWS/2*(cols/8), 
			    u_l2, 
			    ( l_l2 + 
			      (row%(2*M_L2_ROWS))*PANEL_SIZE +
			      (M_L2_ROWS)/2*PANEL_SIZE ),
			    m_l1, 
			    u_l1, 
			    l_l1, 
			    ( ((M_L2_ROWS/(2*BLOCK_SIZE) - 1) << 24) |
			      ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			      (15 << 8) |
			      PANEL_SIZE )); 
#else

#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
    
    /* Read the l_l1 matrix. */
    idma1_setup(l_l1,
		( l_l2 + 
		  (row%(2*M_L2_ROWS))*PANEL_SIZE +
		  (M_L2_ROWS)/2*PANEL_SIZE ), 
		sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
		0, 
		0, 
		7);
    
    /* Read the m_l1 matrix. */
    idma1_setup(m_l1, 
		l2_calc + M_L2_ROWS/2*(cols/8), 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/2, 
		0, 
		0, 
		7);
#endif

    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;

    /* Bring in the first column panel for the U matrix. */
    idma1_setup(u_l1 + PANEL_SIZE*U_L1_COLS, 
		u_l2 + PANEL_SIZE*U_L1_COLS, 
		sizeof(double)*PANEL_SIZE*U_L1_COLS, 
		0, 
		0, 
		7);

    /* Read the m_l1 matrix. */
    idma1_setup(m_l1 + BLOCK_SIZE*BLOCK_SIZE/2, 
		l2_calc + M_L2_ROWS/2*(cols/8) + m_offset, 
		sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/4, 
		0, 
		0, 
		7);


#if 0
    /* Wait until slots are free again. */
    mfence();
    while ((idma1_status() & 0x02) != 0) wait_cnt++;
#endif
    
    wait_cnt +=
      update_block_idma_row(l2_calc + M_L2_ROWS/2*(cols/8), 
			    u_l2, 
			    ( l_l2 + 
			      (row%(2*M_L2_ROWS))*PANEL_SIZE +
			      (M_L2_ROWS)/2*PANEL_SIZE ),
			    m_l1, 
			    u_l1, 
			    l_l1, 
			    ( ((M_L2_ROWS/BLOCK_SIZE - 1) << 24) |
			      ((cols/(8*BLOCK_SIZE) - 1) << 16) |
			      (7 << 8) |
			      PANEL_SIZE )); 
    
#endif
#endif

#if 0
    barrier();
#endif

    if (my_rank == 0) {
      edma_wait_read();
      wait_l_xfer();
    }
    barrier();
      
    tmp = l2_calc; l2_calc = l2_dma; l2_dma = tmp;
  }

    /* Exit: write back results of last row (currently in calc buffer) */
  if (my_rank == 0) {
    for(core=0; core<8;++core) 
      setup_row(core,
		cols/PANEL_SIZE,
		m, 
		m + index_m(row - 64, core*BLOCK_SIZE), 
		l2_dma);
    edma_write();
  }

  wait_cnt +=
    update_block_idma(l2_calc, l2_calc, 
		      u_l2, 
		      l_l2 + (row%(2*M_L2_ROWS))*PANEL_SIZE, 
		      m_l1, u_l1, l_l1, 
		      M_L2_ROWS, cols/8);

  barrier();

  if (my_rank == 0) {
    edma_wait_write();
    
    for(core=0; core<8;++core) 
      setup_row(core,
		cols/PANEL_SIZE,
		m, 
		m + index_m(row, core*BLOCK_SIZE), 
		l2_calc);
    edma_write();
    edma_wait_write();
  }
}

