/**
   Simple linpack block update 
   Calculates m_new = m_old - l*u
   Uses double buffering.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <idma.h>

#if 0
#define PANEL_SIZE 128
#define BLOCK_SIZE 16
#define SUBBLOCK_SIZE 4
#endif

#define L_L1_ROWS 8
#define U_L1_COLS 4

#define USE_IDMA

#if 0
void block_update(double *mnew, 
		  const double *mold,
                  const double *l, 
		  const double *u,
                  unsigned block_count,
		  unsigned panel_count);
#endif

unsigned
update_block_idma(double * restrict m_l2_new,
		  const double * restrict m_l2_old,
		  const double * restrict u_l2,
		  const double * restrict l_l2,
		  double * restrict m_l1,
		  double * restrict u_l1,
		  double * restrict l_l1,
		  unsigned block_rows,
		  unsigned block_cols)
{
  unsigned wait_cnt;
  unsigned block_row, block_col;
  const double *m_l2_load, *l_l2_load;
  double *m_l1_calc, *m_l1_load, *tmp;
  int last_column;
  int u_l2_adjust;
  
  wait_cnt = 0;
  m_l2_load = m_l2_old + BLOCK_SIZE*BLOCK_SIZE;
  m_l1_calc = m_l1;
  m_l1_load = m_l1 + BLOCK_SIZE*BLOCK_SIZE;

  /* Wait until IDMA is free */
  mfence();
  while (!idma1_done()) wait_cnt++;
  
  /* Read the m_l1 matrix. */
  idma1_setup(m_l1_calc, 
	      m_l2_old, 
	      sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 
	      0, 
	      0, 
	      7);

  /* Bring in the first column panel for the U matrix. */
  idma1_setup(u_l1, 
	      u_l2,  
	      sizeof(double)*PANEL_SIZE*U_L1_COLS, 
	      0, 
	      0, 
	      7);

  /* Initialize the L1 buffers. */
  /* Wait until IDMA is free */
  mfence();
  while (!idma1_done()) wait_cnt++;

  /* Read the l_l1 matrix. */
  idma1_setup(l_l1, 
	      l_l2, 
	      sizeof(double)*L_L1_ROWS*PANEL_SIZE, 
	      0, 
	      0, 
	      7);

  for(block_row=0; block_row<block_rows; block_row += BLOCK_SIZE) {
    for(block_col=0; block_col<block_cols; block_col += BLOCK_SIZE) {
      last_column = (block_col + BLOCK_SIZE) >= block_cols;
      if (last_column) {
	u_l2_adjust = -(int)(block_cols*PANEL_SIZE*sizeof(double));
	l_l2_load = l_l2 + index_l_l2(block_row + BLOCK_SIZE, 0);
      } else {
	u_l2_adjust = 0;
	l_l2_load = l_l2 + index_l_l2(block_row, 0);
      }

      /* Update one block*/
      update_block_l2(m_l2_new + index_m_l2(block_row, block_col, block_cols),
		      m_l1_calc,
		      m_l1_load,
		      l_l1,
		      l_l2_load,
		      u_l1,
		      m_l2_load,
		      l_l2 + index_l_l2(block_row + L_L1_ROWS, 0),
		      u_l2 + index_u_l2(0, block_col + SUBBLOCK_SIZE),
		      PANEL_SIZE,
		      u_l2_adjust); 
      m_l2_load += BLOCK_SIZE*BLOCK_SIZE;

 #if 0
      /* Update lower half of block. */
      update_block_l2(( m_l2_new + 
			index_m_l2(block_row + 8, block_col, block_cols) ),
		      m_l1_calc + BLOCK_SIZE*BLOCK_SIZE/2,
		      m_l1_load + BLOCK_SIZE*BLOCK_SIZE/2,
		      l_l1 + PANEL_SIZE*L_L1_ROWS,
		      l_l1,
		      u_l1,
		      m_l2_load,
		      l_l2_load,
		      u_l2 + index_u_l2(0, block_col + SUBBLOCK_SIZE),
		      PANEL_SIZE,
		      u_l2_adjust); 
      m_l2_load += BLOCK_SIZE*BLOCK_SIZE/2;
#endif

      tmp = m_l1_calc; 
      m_l1_calc = m_l1_load;
      m_l1_load = tmp;
    }
  }

  return wait_cnt;
}


#if 0
unsigned
update_block_idma(double * restrict m_l2_new,
		  const double * restrict m_l2_old,
		  const double * restrict u_l2,
		  const double * restrict l_l2,
		  double * restrict m_l1,
		  double * restrict u_l1,
		  double * restrict l_l1,
		  unsigned block_rows,
		  unsigned block_cols)
{
  unsigned wait_cnt;
  unsigned block_row, block_col, row, col, block;
  unsigned m_l1_ofs, m_l2_ofs;
  unsigned l_l1_ofs;
  unsigned u_col_next;
  unsigned l_row_next;

  wait_cnt = 0;

  /* Initialize the L1 buffers. */
  /* Wait until IDMA is free */
  while (!idma1_done()) wait_cnt++;
  /* Read the l_l1 matrix. */
  idma1_setup(l_l1, l_l2, sizeof(double)*L_L1_ROWS*PANEL_SIZE, 0, 0, 7);
  /* Read the m_l1 matrix. */
  idma1_setup(m_l1, m_l2_old, sizeof(double)*BLOCK_SIZE*BLOCK_SIZE, 0, 0, 7);
  /* Wait until one slot is free again. */
  while ((idma1_status() & 0x02) != 0) wait_cnt++;
  /* Bring in the first column panel for the U matrix. */
  idma1_setup(u_l1, u_l2, sizeof(double)*PANEL_SIZE*U_L1_COLS, 0, 0, 7);


  for(block_row=0; block_row<block_rows; block_row += BLOCK_SIZE) {
    for(block_col=0; block_col<block_cols; block_col += BLOCK_SIZE) {

      /* FIXME: need a more effective expression... */
      m_l2_ofs = block_row*block_cols + (block_col + BLOCK_SIZE)*BLOCK_SIZE;
      m_l2_ofs = m_l2_ofs%(block_rows*block_cols);

      for(row=0; row<BLOCK_SIZE; row += L_L1_ROWS) {

	if (row >= (BLOCK_SIZE - L_L1_ROWS)) { /* Last row n block */
	  if (block_col >= (block_cols - BLOCK_SIZE)) { /* Last block in row */
	    l_row_next = block_row + BLOCK_SIZE;
	  } else {
	    l_row_next = block_row;
	  }
	} else {
	  l_row_next = block_row + row + L_L1_ROWS;
	}

	for(col=0; col<BLOCK_SIZE; col += U_L1_COLS) {

	  /* Wait until last transfer is done. */
	  while (!idma1_done()) wait_cnt++;

	  /* Bring in the next column panel for the U matrix. */
	  if (col >= BLOCK_SIZE-U_L1_COLS) { /* Last column in block */
	    if (row >= BLOCK_SIZE-L_L1_ROWS) { /* Last row in block */
	      if (block_col >= block_cols-BLOCK_SIZE) { /* Last column in row */
		u_col_next = 0;
	      } else {
		u_col_next = block_col + col + U_L1_COLS;
	      }
	    } else {
	      u_col_next = block_col;
	    }
	  } else {
	    u_col_next = block_col + col + U_L1_COLS;
	  }

	  idma1_setup(u_l1 + (u_col_next%(2*U_L1_COLS))*PANEL_SIZE,
		      u_l2 + u_col_next*PANEL_SIZE,
		      sizeof(double)*PANEL_SIZE*U_L1_COLS,
		      0, 
		      0, 
		      7);
		      
	  block = col/U_L1_COLS;
	  l_l1_ofs = (block - 2)*L_L1_ROWS*PANEL_SIZE/2;
	  m_l1_ofs = row*BLOCK_SIZE + block*BLOCK_SIZE*BLOCK_SIZE/4;
	  
	  if (block < 2) {
	    /* Read 1/4 of the m_l1 matrix. */
	    idma1_setup(m_l1 + m_l1_ofs + m_l2_ofs%(2*BLOCK_SIZE*BLOCK_SIZE),
			m_l2_old + m_l1_ofs + m_l2_ofs,
			sizeof(double)*BLOCK_SIZE*BLOCK_SIZE/4,
			0,
			0,
			7);
	  } else {
	    /* Read half of the l_l1 matrix. */
	    idma1_setup(l_l1 + (l_row_next%(2*L_L1_ROWS))*PANEL_SIZE + l_l1_ofs,
			l_l2 + l_row_next*PANEL_SIZE + l_l1_ofs,
			sizeof(double)*L_L1_ROWS*PANEL_SIZE/2,
			0,
			0,
			7);
	  }

#if 0  
	  block_update(( m_l2_new + 
			 block_row*block_cols + block_col*BLOCK_SIZE + 
			 row*BLOCK_SIZE + col ),
		       ( m_l1 + 
			 ( (block_row*block_cols + block_col*BLOCK_SIZE)%
			   (2*BLOCK_SIZE*BLOCK_SIZE) ) +
			 row*BLOCK_SIZE + col ),
		       l_l1 + row*PANEL_SIZE,
		       u_l1 + (col%(2*U_L1_COLS))*PANEL_SIZE,
		       L_L1_ROWS/SUBBLOCK_SIZE,
		       PANEL_SIZE);
#else
	  update_block_l2(( m_l2_new + 
			    block_row*block_cols + block_col*BLOCK_SIZE + 
			    row*BLOCK_SIZE + col ),
			  ( m_l1 + 
			    ( (block_row*block_cols + block_col*BLOCK_SIZE)%
			      (2*BLOCK_SIZE*BLOCK_SIZE) ) +
			    row*BLOCK_SIZE + col ),
			  0,
			  l_l1 + row*PANEL_SIZE,
			  0,
			  u_l1 + (col%(2*U_L1_COLS))*PANEL_SIZE,
			  0,
			  0,
			  0,
			  PANEL_SIZE,
			  0);
#endif

	}
      }
      
    }
  }

  return wait_cnt;
}
#endif
