/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <cache.h>
#include <barrier.h>
#include <tsc.h>

#include <stdio.h>

/* The shared (MCSM) buffers. */
#pragma DATA_SECTION(L_l3, "panel")
double L_l3[2*M_L2_ROWS*PANEL_SIZE];

/* The L2 cache buffers. */
#pragma DATA_SECTION(M_l2, "l2")
double M_l2[2*M_L2_ROWS*M_L2_COLS];
#pragma DATA_SECTION(U_l2, "l2")
double U_l2[2*PANEL_SIZE*M_L2_COLS];
#pragma DATA_SECTION(L_l2, "l2")
double L_l2[2*M_L2_ROWS*PANEL_SIZE];

/* The L1 cache buffers. */
#pragma DATA_SECTION(L_l1, "l1d")
double L_l1[2*L_L1_ROWS*PANEL_SIZE];
#pragma DATA_SECTION(U_l1, "l1d")
double U_l1[2*PANEL_SIZE*U_L1_COLS];
#pragma DATA_SECTION(M_l1, "l1d")
double M_l1[2*BLOCK_SIZE*BLOCK_SIZE];

#pragma DATA_SECTION(l_tmp, "panel");
static double l_tmp[PANEL_SIZE*PANEL_SIZE];

void
setup_m_panel()
{
  unsigned my_rank;
  
  my_rank = cache_get_mm_id();
  if (my_rank == 0) {
    update_block_setup_edma(M_l2, L_l2, L_l3);
  }
  barrier();
}

uint64_t 
update_m_panel(double * restrict m,
	       const size_t * restrict pivot_rows,
	       size_t panel_width, 
	       size_t cols)
{
  size_t col;
  unsigned my_rank;
  size_t m_panel_width;
#if 0
  uint64_t start_t, end_t, accum_t, accum_u_t, accum_c_t;
  double time, flop;
  
  accum_t = 0;
  accum_u_t = 0;
  accum_c_t = 0;
#endif

  my_rank = cache_get_mm_id();

#if 0
  start_t = tsc_get();

  if (my_rank == 0) {
    transpose_l_panel(m + index_m(panel_width, 0),
		      panel_width,
		      cols - panel_width);
  }
  coherence_flush_all();
  barrier();

  end_t = tsc_get();
  if (my_rank == 0) {
    printf("transpose_l_panel took %llu cycles for %d rows\n",
	   end_t - start_t,
	   cols - panel_width);
  }
#endif

  for(col=panel_width; col<cols; col+=8*PANEL_SIZE) {
    m_panel_width = ((col + 8*PANEL_SIZE)>cols)?(cols - col):(8*PANEL_SIZE);

#if 0
    start_t = tsc_get();
#endif
    
    fill_u_panel_l2(U_l2,
		    m + index_m(0, col + my_rank*BLOCK_SIZE),
		    pivot_rows,
		    PANEL_SIZE,
		    m_panel_width);
    fill_l_head_l2(L_l2,
		   m,
		   PANEL_SIZE,
		   PANEL_SIZE);
    
#if 0
    end_t = tsc_get();
    accum_c_t += end_t - start_t;
		   
    start_t = tsc_get();
#endif
    
    update_u_panel_l2(U_l2, L_l2, PANEL_SIZE, m_panel_width/8);

#if 0		      
    end_t = tsc_get();
    accum_u_t += end_t - start_t;

    start_t = tsc_get();
#endif

    save_u_panel_l2(m + index_m(0, col + my_rank*BLOCK_SIZE),
		    U_l2,
		    PANEL_SIZE,
		    m_panel_width);

#if 0
    end_t = tsc_get();
    accum_c_t += end_t - start_t;
#endif

    cache_l1d_set_size(1);	/* 4kB L1D cache, 28kB SRAM */
    cache_l2_set_size(0);	/* 0kB L2, 512kB SRAM */ 

#if 0
    start_t = tsc_get();
#endif
    update_block_edma(m + index_m(panel_width, col),
		      m + index_m(panel_width, 0),
		      L_l3,
		      M_l2,
		      U_l2,
		      L_l2,
		      M_l1,
		      U_l1,
		      L_l1,
		      cols - panel_width,
		      m_panel_width);
#if 0
    end_t = tsc_get();
    accum_t += end_t - start_t;
#endif

    cache_l1d_set_size(4);	/* 32kB L1D cache, 0kB SRAM */
    cache_l2_set_size(0);	/* 0kB L2, 512kB SRAM */ 
    
  }

#if 0
  if (my_rank == 0) {
    time = (double)(accum_u_t);
    flop = (double)(cols - PANEL_SIZE)*PANEL_SIZE*(PANEL_SIZE + 1);
    printf("update_u_panel took %llu cycles for %u cols (=%6.4f FPC)" 
	   "copy took %llu cycles\n",
	   accum_u_t, 
	   cols - PANEL_SIZE,
	   flop/time,
	   accum_c_t);
  }
#endif

#if 0
  return accum_t;
#else
  return 0;
#endif
}

void
transpose_l_panel(double * restrict m,
		  size_t panel_width,
		  size_t rows)
{
  size_t row;
  size_t block_row, block_col;

  for(row=0; row<rows; row+=panel_width) {
    for(block_row=0; block_row<panel_width; ++block_row) {
      for(block_col=0; block_col<panel_width; ++block_col) {
	l_tmp[block_row*panel_width + block_col] = 
	  m[index_m(row + block_row, block_col)];
      }
    }

    for(block_row=0; block_row<panel_width; ++block_row) {
      for(block_col=0; block_col<panel_width; ++block_col) {
	m[index_l(row + block_row, block_col)] = 
	  l_tmp[block_row*panel_width + block_col];
      }
    }
  }
}



