/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include <stdio.h>
#include <math.h>

#include <tsc.h>
#include <barrier.h>
#include <cache.h>
#include <time.h>
#include <log.h>

#include "coherence.h"
#include "index.h"

#define NMAX 8175

/* Number of rows in a panel (MUST BE multiple of 4!). */
#define PANEL_ROWS 32

#define BLOCK_COUNT (PANEL_ROWS/4)

/* Static Data declaration, Fortan style. */

/* M holds the matrix [ A b ], therefore N rows +1, N+1 columns */
#pragma DATA_SECTION(m, "matrix")
double m[(NMAX + 1)*(NMAX + 1)];

#define AIDX(row, col) ((row)*(NMAX + 1) + (col))
#define BIDX(row) ((row)*(NMAX + 1) + n)

/* u holds the upper panel (rows), l holds the lower panel (columns) */
#pragma DATA_SECTION(u, "panel")
double u[(NMAX + 1)*PANEL_ROWS];
#pragma DATA_SECTION(l, "panel");
double l[(NMAX + 1)*PANEL_ROWS];

#define PANEL_SIZE ((NMAX + 1)*PANEL_ROWS)

#define UIDX(row, col) ((row)*(NMAX + 1) + (col))
#define LIDX(row, col) ((col)*(NMAX + 1) + (row))

/* The solution vector. */
#pragma DATA_SECTION(x, "matrix")
double x[NMAX];

/* Residual vector. */
#pragma DATA_SECTION(r, "matrix")
double r[NMAX];

int random_seed = 1325;
unsigned n = NMAX;

static const unsigned long long idle_length = 5000000000ull;

#pragma FUNC_CANNOT_INLINE(init_matrix);
static void
init_matrix()
{
  unsigned i, j;

  for(i=0; i<n; i++) {
    for(j=0; j<n; j++) {
      random_seed = (int)((long)3125*(long)random_seed % 65536L);
      m[AIDX(i,j)] = (random_seed - 32768.0)/16384.0;
    }
  }
}

#pragma FUNC_CANNOT_INLINE(init_problem);
static void
init_problem()
{
  unsigned i,j;

  random_seed = 1325;
  init_matrix();

  for (i = 0; i < n; i++) 
    m[BIDX(i)] = 0.0;

  for (j = 0; j < n; j++)
    for (i = 0; i < n; i++)
      m[BIDX(i)] += m[AIDX(j,i)];

#if 0
  for(i=0; i<(n+1); i++) {
    for(j=0; j<(n+1); j++) {
      m[AIDX(i,j)] = 100*i+j;
    }
  }
#endif
}

static void
print_matrix()
{
  printf("m:\n");
  unsigned i,j;
  for(i=0; i<n; ++i) {
    for(j=0; j<(n+1); ++j) printf("%+7.3e ", m[AIDX(i,j)]);
    printf("\n");
  }
}

/* Copies the panel starting in given column to L (with proper row offset). */
/* DMA candidate! */
static void
fill_lower_panel(unsigned start_col)
{
  unsigned row, col;

  for(row=start_col; row<n; ++row) {
    for(col=0; col<PANEL_ROWS; ++col) {
      l[LIDX(row, col)] = m[AIDX(row, start_col + col)];
    }
  }

  for(col=0; col<PANEL_ROWS; ++col) l[LIDX(n, col)] = 0.0;
}

/* Copies the right part of the pivoted row into the upper panel. */
/* DMA candidate! */
static void
fill_upper_panel(unsigned start_col, 
		 unsigned panel_row, 
		 unsigned pivot_row)
{
  unsigned col, end_col, start_ofs;
  double tmp;

  end_col = n + 1 - start_col + PANEL_ROWS;
  start_ofs = start_col - PANEL_ROWS;

  for(col=PANEL_ROWS; col<end_col; ++col) {
    tmp = u[UIDX(panel_row, col)];
    u[UIDX(panel_row, col)] = m[AIDX(pivot_row, col + start_ofs)];
    m[AIDX(pivot_row, col + start_ofs)] = tmp;
  }
}

/* Copies the right part of the pivoted row into the upper panel. */
/* DMA candidate! */
static void
fill_upper_panel_u(unsigned start_col, 
		   unsigned panel_row, 
		   unsigned pivot_row)
{
  unsigned col, end_col;
  double tmp;

  end_col = n + 1 - start_col + PANEL_ROWS;

  for(col=PANEL_ROWS; col<end_col; ++col) {
    tmp = u[UIDX(panel_row, col)];
    u[UIDX(panel_row, col)] = u[UIDX(pivot_row, col)];
    u[UIDX(pivot_row, col)] = tmp;
  }
}

/* Save back the top part of the lower panel to m. */
static void
save_lower_panel(unsigned start_col)
{
  unsigned row, col;
  for(row=0; row<PANEL_ROWS; ++row) {
    for(col=0; col<PANEL_ROWS; ++col) {
      m[AIDX(start_col + row, start_col + col)] = l[LIDX(row, col)];
    }
  }
}

/* Save back the upper panel to m. */
static void
save_upper_panel(unsigned start_col) 
{
  unsigned row, col;
  for(row=0; row<PANEL_ROWS; ++row) {
    for(col=PANEL_ROWS; col<(n + 1 - start_col); ++col) {
      m[AIDX(row + start_col, col + start_col)] = u[UIDX(row, col)];
    }
  }
}

/* Find the pivot row in the lower panel. */
static unsigned
find_pivot_row(unsigned start_col, unsigned panel_col)
{
  unsigned row;
  unsigned pivot_row, end_row;
  double pivot_abs, cur_abs;

  end_row = n - start_col;
  pivot_row = panel_col;
  pivot_abs = fabs(l[LIDX(pivot_row, panel_col)]);
  
  for(row = pivot_row + 1; row<end_row; ++row) {
    cur_abs = fabs(l[LIDX(row, panel_col)]);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }

  return pivot_row;
}

/* Update a block based on the upper and lower panel. */
static void
update_block(unsigned row, unsigned col, unsigned start_col) 
{
  double * restrict lp, * restrict up;
  double * restrict mp;
  int i, j, k;

  double ll0, ll1, ll2, ll3;
  double ul0, ul1, ul2, ul3;

  double ml00, ml01, ml02, ml03;
  double ml10, ml11, ml12, ml13;
  double ml20, ml21, ml22, ml23;
  double ml30, ml31, ml32, ml33;

  up = u + col - start_col;
  lp = l + row - start_col;
  mp = m + row*(NMAX + 1) + col;

  for(i=BLOCK_COUNT - 1; i>=0; --i) {

    for(j=BLOCK_COUNT - 1; j>=0; --j) {

      ul0 = *(up++);
      ul1 = *(up++);
      ul2 = *(up++);
      ul3 = *(up); 
      up += NMAX + 1 - 3;
      
      ll0 = *(lp++);
      ll1 = *(lp++);
      ll2 = *(lp++);
      ll3 = *(lp); 
      lp += NMAX + 1 - 3;
      
      ml00 = ll0*ul0;
      ml01 = ll0*ul1;
      ml02 = ll0*ul2;
      ml03 = ll0*ul3;
      
      ml10 = ll1*ul0;
      ml11 = ll1*ul1;
      ml12 = ll1*ul2;
      ml13 = ll1*ul3;
      
      ml20 = ll2*ul0;
      ml21 = ll2*ul1;
      ml22 = ll2*ul2;
      ml23 = ll2*ul3;
      
      ml30 = ll3*ul0;
      ml31 = ll3*ul1;
      ml32 = ll3*ul2;
      ml33 = ll3*ul3;

      for(k=PANEL_ROWS - 2; k>=0; --k) {
	ul0 = *(up++);
	ul1 = *(up++);
	ul2 = *(up++);
	ul3 = *(up); 
	up += NMAX + 1 - 3;

	ll0 = *(lp++);
	ll1 = *(lp++);
	ll2 = *(lp++);
	ll3 = *(lp); 
	lp += NMAX + 1 - 3;

	ml00 += ll0*ul0;
	ml01 += ll0*ul1;
	ml02 += ll0*ul2;
	ml03 += ll0*ul3;

	ml10 += ll1*ul0;
	ml11 += ll1*ul1;
	ml12 += ll1*ul2;
	ml13 += ll1*ul3;

	ml20 += ll2*ul0;
	ml21 += ll2*ul1;
	ml22 += ll2*ul2;
	ml23 += ll2*ul3;

	ml30 += ll3*ul0;
	ml31 += ll3*ul1;
	ml32 += ll3*ul2;
	ml33 += ll3*ul3;
      } /* k */

      *(mp++) = *(mp) - ml00;
      *(mp++) = *(mp) - ml01;
      *(mp++) = *(mp) - ml02;
      *(mp) = *(mp) - ml03;
      mp += NMAX + 1 - 3;

      *(mp++) = *(mp) - ml10;
      *(mp++) = *(mp) - ml11;
      *(mp++) = *(mp) - ml12;
      *(mp) = *(mp) - ml13;
       mp += NMAX + 1 - 3;

      *(mp++) = *(mp) - ml20;
      *(mp++) = *(mp) - ml21;
      *(mp++) = *(mp) - ml22;
      *(mp) = *(mp) - ml23;
      mp += NMAX + 1 - 3;

      *(mp++) = *(mp) - ml30;
      *(mp++) = *(mp) - ml31;
      *(mp++) = *(mp) - ml32;
      *(mp) = *(mp) - ml33;
      mp -= (3*(NMAX + 1) - 1);

      lp -= PANEL_ROWS*(NMAX + 1);
      up -= PANEL_ROWS*(NMAX + 1) - 4;
    } /* j */

    lp += 4;
    up -= PANEL_ROWS;
    mp += 4*(NMAX + 1) - PANEL_ROWS;
  } /* i */
}

/* Update a block based on the upper and lower panel. write result into u */
static void
update_block_u(unsigned row, unsigned col, unsigned start_col) 
{
  double * lp, *up, *mp, *unp;
  int i, j, k;

  double ll0, ll1, ll2, ll3;
  double ul0, ul1, ul2, ul3;

  double ml00, ml01, ml02, ml03;
  double ml10, ml11, ml12, ml13;
  double ml20, ml21, ml22, ml23;
  double ml30, ml31, ml32, ml33;

  up = u + col - start_col;
  lp = l + PANEL_ROWS;
  mp = m + row*(NMAX + 1) + col;
  unp = u + col - start_col - PANEL_ROWS;


  for(i=BLOCK_COUNT - 1; i>=0; --i) {

    for(j=BLOCK_COUNT - 1; j>=0; --j) {

      ml00 = mp[ 0];
      ml01 = mp[ 1];
      ml02 = mp[ 2];
      ml03 = mp[ 3];
      mp += NMAX + 1;
  
      ml10 = mp[ 0];
      ml11 = mp[ 1];
      ml12 = mp[ 2];
      ml13 = mp[ 3];
      mp += NMAX + 1;

      ml20 = mp[ 0];
      ml21 = mp[ 1];
      ml22 = mp[ 2];
      ml23 = mp[ 3];
      mp += NMAX + 1;

      ml30 = mp[ 0];
      ml31 = mp[ 1];
      ml32 = mp[ 2];
      ml33 = mp[ 3];
      mp -= (3*(NMAX + 1) - 4);

      for(k=PANEL_ROWS - 1; k>=0; --k) {
	ul0 = up[0];
	ul1 = up[1];
	ul2 = up[2];
	ul3 = up[3];
	up += NMAX + 1;

	ll0 = lp[0];

	ml00 -= ll0*ul0;
	ml01 -= ll0*ul1;
	ml02 -= ll0*ul2;
	ml03 -= ll0*ul3;

	ll1 = lp[1];

	ml10 -= ll1*ul0;
	ml11 -= ll1*ul1;
	ml12 -= ll1*ul2;
	ml13 -= ll1*ul3;

	ll2 = lp[2];

	ml20 -= ll2*ul0;
	ml21 -= ll2*ul1;
	ml22 -= ll2*ul2;
	ml23 -= ll2*ul3;

	ll3 = lp[3];

	ml30 -= ll3*ul0;
	ml31 -= ll3*ul1;
	ml32 -= ll3*ul2;
	ml33 -= ll3*ul3;

	lp += NMAX + 1;
      } /* k */

      unp[ 0] = ml00;
      unp[ 1] = ml01;
      unp[ 2] = ml02;
      unp[ 3] = ml03;
      unp += NMAX + 1;

      unp[ 0] = ml10;
      unp[ 1] = ml11;
      unp[ 2] = ml12;
      unp[ 3] = ml13;
      unp += NMAX + 1;

      unp[ 0] = ml20;
      unp[ 1] = ml21;
      unp[ 2] = ml22;
      unp[ 3] = ml23;
      unp += NMAX + 1;

      unp[ 0] = ml30;
      unp[ 1] = ml31;
      unp[ 2] = ml32;
      unp[ 3] = ml33;
      unp -= (3*(NMAX + 1) - 4);

      lp -= PANEL_ROWS*(NMAX + 1);
      up -= PANEL_ROWS*(NMAX + 1) - 4;
    } /* j */

    lp += 4;
    up -= PANEL_ROWS;
    mp += 4*(NMAX + 1) - PANEL_ROWS;
    unp += 4*(NMAX + 1) - PANEL_ROWS;
  } /* i */
}

/* Update a block based on the upper and lower panel. write result into l */
static void
update_block_l(unsigned row, unsigned col, unsigned start_row) 
{
  double * lp, *up, *mp, *lnp;
  int i, j, k;

  double ll0, ll1, ll2, ll3;
  double ul0, ul1, ul2, ul3;

  double ml00, ml01, ml02, ml03;
  double ml10, ml11, ml12, ml13;
  double ml20, ml21, ml22, ml23;
  double ml30, ml31, ml32, ml33;

  up = u;
  lp = l + row - start_row;
  mp = m + row*(NMAX + 1) + col;
  lnp = l + row - start_row - PANEL_ROWS;

  for(i=BLOCK_COUNT - 1; i>=0; --i) {

    for(j=BLOCK_COUNT - 1; j>=0; --j) {

      ml00 = mp[ 0];
      ml01 = mp[ 1];
      ml02 = mp[ 2];
      ml03 = mp[ 3];
      mp += NMAX + 1;
  
      ml10 = mp[ 0];
      ml11 = mp[ 1];
      ml12 = mp[ 2];
      ml13 = mp[ 3];
      mp += NMAX + 1;

      ml20 = mp[ 0];
      ml21 = mp[ 1];
      ml22 = mp[ 2];
      ml23 = mp[ 3];
      mp += NMAX + 1;

      ml30 = mp[ 0];
      ml31 = mp[ 1];
      ml32 = mp[ 2];
      ml33 = mp[ 3];
      mp -= (3*(NMAX + 1) - 4);

      for(k=PANEL_ROWS - 1; k>=0; --k) {
	ul0 = up[0];
	ul1 = up[1];
	ul2 = up[2];
	ul3 = up[3];
	up += NMAX + 1;

	ll0 = lp[0];

	ml00 -= ll0*ul0;
	ml01 -= ll0*ul1;
	ml02 -= ll0*ul2;
	ml03 -= ll0*ul3;

	ll1 = lp[1];

	ml10 -= ll1*ul0;
	ml11 -= ll1*ul1;
	ml12 -= ll1*ul2;
	ml13 -= ll1*ul3;

	ll2 = lp[2];

	ml20 -= ll2*ul0;
	ml21 -= ll2*ul1;
	ml22 -= ll2*ul2;
	ml23 -= ll2*ul3;

	ll3 = lp[3];

	ml30 -= ll3*ul0;
	ml31 -= ll3*ul1;
	ml32 -= ll3*ul2;
	ml33 -= ll3*ul3;

	lp += NMAX + 1;
      } /* k */

      lnp[ 0] = ml00;
      lnp[ 1] = ml10;
      lnp[ 2] = ml20;
      lnp[ 3] = ml30;
      lnp += NMAX + 1;

      lnp[ 0] = ml01;
      lnp[ 1] = ml11;
      lnp[ 2] = ml21;
      lnp[ 3] = ml31;
      lnp += NMAX + 1;

      lnp[ 0] = ml02;
      lnp[ 1] = ml12;
      lnp[ 2] = ml22;
      lnp[ 3] = ml32;
      lnp += NMAX + 1;

      lnp[ 0] = ml03;
      lnp[ 1] = ml13;
      lnp[ 2] = ml23;
      lnp[ 3] = ml33;
      lnp += NMAX + 1;

      lp -= PANEL_ROWS*(NMAX + 1);
      up -= PANEL_ROWS*(NMAX + 1) - 4;
    } /* j */

    lp += 4;
    up -= PANEL_ROWS;
    mp += 4*(NMAX + 1) - PANEL_ROWS;
    lnp -= PANEL_ROWS*(NMAX + 1) - 4;
  } /* i */
}

static void
factorize()
{
  unsigned start_col, panel_col, next_start_col, panel_end, panel_top;
  unsigned panel_row, panel_end_row;
  unsigned row, col;
  unsigned block_start_col;

  unsigned pivot_row;
  double pivot_scale, row_scale;
  double tmp, saved_top, sum;
  double *p_tmp;

  unsigned my_rank, num_ranks;

  unsigned first_block, last_block, num_blocks, num_cols;
  unsigned first_row, last_row, first_col, last_col;

  uint64_t start_time, end_time;

  my_rank = cache_get_mm_id();
  num_ranks = cache_get_mm_num();

  if (my_rank == 0) {		/* Master only */
    fill_lower_panel(0);	/* Initialize L panel. */
  
    /* Initialize the U panel. */
    for(row=0; row<PANEL_ROWS; row++) {
      for(col=PANEL_ROWS; col<(n + 1); col++) {
	u[UIDX(row, col)] = m[AIDX(row, col)];
      }
    }
  }
  barrier();

#if 0
  print_matrix();
#endif

  /* We first eliminate everything in panels of columns. */
  for(start_col=0; start_col<(n + 1); start_col = next_start_col) {
    next_start_col = start_col + PANEL_ROWS;
    block_start_col = next_start_col + PANEL_ROWS;

    if (my_rank == 0) {		/* Master only! */

      /* Last panel is one column short (b) */
      panel_end = (next_start_col>n)?(PANEL_ROWS - 1):PANEL_ROWS;

      /* Eliminate each column in the panel. */
      for(panel_col=0; panel_col<panel_end; panel_col++) {
      
	panel_top = start_col + panel_col; /* Top row in the panel */
	panel_end_row = n - start_col;	 /* Last row in the panel */

	pivot_row = find_pivot_row(start_col, panel_col);
	saved_top = l[LIDX(pivot_row, panel_col)];
	pivot_scale = 1.0/saved_top;

	/* Scale pivot row and exchange with top of panel row. */
	if (pivot_row != 0) {
	  for(col = 0; col<PANEL_ROWS; col++) {
	    tmp = l[LIDX(panel_col, col)];
	    l[LIDX(panel_col, col)] = l[LIDX(pivot_row, col)];
	    l[LIDX(pivot_row, col)] = tmp;
	  }
	}
	for(col = panel_col + 1; col<PANEL_ROWS; col++) {
	  l[LIDX(panel_col, col)] *= pivot_scale;
	}
	l[LIDX(panel_col, panel_col)] = pivot_scale;
      
      
	/* Eliminate pivot row from l panel, leave the scale factors. */
	for(row = panel_col + 1; row<panel_end_row; row++) {
	  row_scale = l[LIDX(row, panel_col)];
	  for(col = panel_col + 1; col<PANEL_ROWS; col++) {
	    l[LIDX(row, col)] -= row_scale*l[LIDX(panel_col, col)];
	  }
	}

	/* Exchange pivot row into upper panel. */
	if (pivot_row < PANEL_ROWS) {
	  fill_upper_panel_u(next_start_col, panel_col, pivot_row);
	} else {
	  fill_upper_panel(next_start_col, panel_col, pivot_row + start_col);
	}

      } /* panel_col */

      /* Eliminate equations in the upper panel, scale at the same time. */
      for(panel_row = 0; panel_row<panel_end; ++panel_row) {
	panel_top = start_col + panel_row;
	row_scale = l[LIDX(panel_row, panel_row)];
      
	for(col = PANEL_ROWS; col<(n + 1 - start_col); ++col) {
	  sum = 0.0;
	  for(row = 0; row < panel_row; ++row) {
	    sum += l[LIDX(panel_row, row)]*u[UIDX(row, col)];
	  } 

	  u[UIDX(panel_row, col)] = row_scale*(u[UIDX(panel_row, col)] - sum);
	}	/* col */
      } /* panel_row */
      for(panel_row = panel_end; panel_row<PANEL_ROWS; panel_row++) {
	for(col = PANEL_ROWS; col<(n + 1 - start_col); ++col) {
	  u[UIDX(row, col)]=0.0;
	}
      }

      /* Save back the factorized parts of the panels to m. */
      save_lower_panel(start_col);
      if (next_start_col < (n + 1)) save_upper_panel(start_col);

      /* Blocked update the matrix m. */
      if (next_start_col < (n + 1)) {

	/* Copy away the first U block for the L update. */
	for(row = 0; row<PANEL_ROWS; row++) {
	  for(col = 0; col<PANEL_ROWS; col++) {
	    u[UIDX(row, col)] = u[UIDX(row, col + PANEL_ROWS)];
	  }
	}
      }
    } /* Master only */

    coherence_flush_all();
    barrier();

    start_time = tsc_get();
    
    /* Now update the remaining parts of M (if any). */   
    if (next_start_col < (n + 1)) {
      if (block_start_col<(n + 1)) {

	/* Firgure out our part of the work. */
	num_cols = (n + 1 - block_start_col)/PANEL_ROWS;

#if 0
	index_distribution_static(0,
				  num_cols,
				  num_ranks,
				  my_rank,
				  &first_row,
				  &last_row);

	first_row = first_row*PANEL_ROWS + block_start_col;
	last_row = last_row*PANEL_ROWS + block_start_col;

	for(row = first_row; row<last_row; row+=PANEL_ROWS) {
	  for(col = block_start_col; col<(n + 1); col+=PANEL_ROWS) {
	    update_block(row, col, start_col);
	  }
	}
#endif

	num_blocks = num_cols*num_cols;

	index_distribution_static(0,
				  num_blocks,
				  num_ranks,
				  my_rank,
				  &first_block,
				  &last_block);
	if (last_block > first_block) {
	  
	  first_row = first_block/num_cols;
	  first_col = first_block - first_row*num_cols;
	  last_row = (last_block - 1)/num_cols;
	  last_col = last_block - last_row*num_cols;

	  if (first_row == last_row) { /* Only one row to work on... */
	    row = first_row*PANEL_ROWS + block_start_col;
	    for(col = first_col; col<last_col; ++col) {
	      update_block(row, col*PANEL_ROWS + block_start_col, start_col);
	    }
	  } else {
	    /* first row. */
	    row = first_row*PANEL_ROWS + block_start_col;
	    for(col = first_col*PANEL_ROWS + block_start_col; 
		col<(n + 1); 
		col+=PANEL_ROWS) {
	      update_block(row, col, start_col);
	    }
	    
	    /* in-between rows. */
	    for(row = first_row + 1; row<last_row; ++row) {
	      for(col = block_start_col; col<(n + 1); col+=PANEL_ROWS) {
		update_block(row*PANEL_ROWS + block_start_col, col, start_col);
	      }
	    }
	  
	    /* last row */
	    row = last_row*PANEL_ROWS + block_start_col;
	    for(col = block_start_col; 
		col<(last_col*PANEL_ROWS + block_start_col); 
		col+=PANEL_ROWS) {
	      update_block(row, col, start_col);
	    }
	  }
	}
      }
    }

    coherence_flush_all();
    barrier();

    end_time = tsc_get();

#if 0
    if (my_rank == 0) {
      printf("Block update %u (start_col = %u) took %llu cycles\n",
	     num_blocks,
	     block_start_col,
	     end_time - start_time);
    }
#endif

    if (my_rank == 0) { 	/* Master only */
      if (next_start_col < (n + 1)) {
	if (block_start_col<(n + 1)) {
	  /* Save the top PANEL into the U space. */
	  for(col = block_start_col; col<(n + 1); col+=PANEL_ROWS) {
	    update_block_u(next_start_col, col, start_col);
	  }
	}
	
	/* Update the L panel last. */
	for(row = next_start_col; row<(n + 1); row+=PANEL_ROWS) {
	  update_block_l(row, next_start_col, start_col);
	}
      }
    }
          
#if 0
    print_matrix();
#endif

  } /* start_col */
}


static void
solve()
{
  int i,j;
  double xcur;

  /* Backwards substitution. */
  for(i=n-1; i>=0; i--) {
    xcur = 0.0;
    for(j=n-1; j>i; j--) xcur += m[AIDX(i,j)]*x[j];
    
    x[i] = m[BIDX(i)] - xcur;
  }
}

static int
verify()
{
  const double eps = 2.2e-16;

  unsigned i, j;
  double rcur;
  double rinf = 0.0; 
  double rabs = 0.0;
  double aone = 0.0;
  double ainf = 0.0;
  double xinf = 0.0;
  double xone = 0.0;
  double binf = 0.0;
  double asum;
  double r0, r1, r2, r3;

  /* Recreate original matrix. */
  init_problem();

  /* Calculate residual r = Ax - b and norms (max and euclid) */
  for(i=0; i<n; i++) {
    rcur = 0.0;
    for(j=0; j<n; j++) rcur += m[AIDX(i,j)]*x[j];
    rcur = rcur - m[BIDX(i)];

    rabs += rcur*rcur;
    if (fabs(rcur)>rinf) rinf=fabs(rcur);
  }

  for(i=0; i<n; i++) {
    asum=0.0;
    for(j=0; j<n; j++) asum += fabs(m[AIDX(i,j)]);
    if (asum > ainf) ainf = asum;
  }

  for(i=0; i<n; i++) {
    asum=0.0;
    for(j=0; j<n; j++) asum += fabs(m[AIDX(j,i)]);
    if (asum > aone) aone = asum;
  }

  xinf = 0.0; xone = 0.0;
  for(i=0; i<n; i++) {
    xone += fabs(x[i]);
    if (fabs(x[i])>xinf) xinf = fabs(x[i]);
  }

  binf = 0.0;
  for(i=0; i<n; i++) {
    if (fabs(m[BIDX(i)])>binf) binf = fabs(m[BIDX(i)]);
  }

  printf("rabs : %g  rinf : %g aone : %g ainf %g xinf %g\n", 
	 sqrt(rabs), rinf, aone, ainf, xinf );

  r0 = rinf / (eps*(xinf*ainf + binf)*n);
  r1 = rinf / (eps*aone*n);
  r2 = rinf / (eps*aone*xone);
  r3 = rinf / (eps*ainf*xinf);

  printf("r0 : % g r1 : %g r2 : %g r3 : %g\n", r0, r1, r2, r3);

  return (r0 < 1.0)?0:1;
}

uint64_t 
linpack(unsigned size)
{
  uint64_t start_t, end_t;
  unsigned my_rank;

  my_rank = cache_get_mm_id();
  
  if (my_rank == 0) {
    n=size;
    
    init_problem();
  };

  coherence_flush_all();
  barrier();
  
  start_t = tsc_get();
  
  factorize();
  
  coherence_flush_all();
  barrier();
  
  if (my_rank == 0) {
    solve();
  }
  
  barrier();

  end_t = tsc_get();

  return end_t - start_t;
}

struct param_t {
  long size;
  long repetitions;
};

const struct param_t parameters[] = {
#if 0
  {   127, 260000 },
  {   255,  62000 },
  {   511,  15000 },
  {  1023,   3200 },
  {  2047,    660 },
  {  4095,    130 },
  {  8159,     21 },
#endif
  {   127, 1 },
  {   255, 1 },
  {   511, 1 },
  {  1023, 1 },
  {  2047, 1 },
  {  4095, 1 },
  {  8063, 1 }, 
  {  8159, 1 },
  {    -1,     -1 }
};

int
main(int argc, char **argv)
{
  int ret;
  uint64_t start_t, end_t, acc_t;
  double sol_t, sol_ops, sol_acc;
  uint64_t bu_t;
  unsigned rep;
  
  unsigned my_rank, num_ranks;
  unsigned size, nrep, warm_rep;

  const struct param_t *cur_run = parameters;

  /* Mark the DDR as cacheable */
  cache_set_memory_attributes((void*)(0x80000000), 
			      0x7fffffffull, 
			      CACHE_MAR_CACHEABLE);

  my_rank = cache_get_mm_id();
  num_ranks = cache_get_mm_num();

  while (cur_run->size > 0) {
    size = cur_run->size;
    nrep = cur_run->repetitions;
    warm_rep = (nrep + 9)/10;
    cur_run++;

    acc_t = 0;

    if (my_rank == 0) {
      log_write_name("linpack");
      log_write_param_uint64("size", size);
      log_write_param_uint64("iterations", nrep);
      log_write_param_uint64("ranks", num_ranks);
    }

    barrier();

#if 0
    /* Write idle shoulder. */
    if (my_rank == 0) log_write_marker("idle_start");
    nanosleep(idle_length);
    barrier();
    
    /* Output warmup marker. */
    if (my_rank == 0) {
      log_write_marker("run_start");
    }
    barrier();

    for(rep=0; rep<warm_rep; ++rep) {
      linpack(size);
    }
#endif

    barrier();
    if (my_rank == 0) log_write_start();
    start_t = tsc_get();

    for(rep=0; rep<nrep; ++rep) {
      acc_t += linpack(size);
    }

    barrier();
    end_t = tsc_get();
    if (my_rank == 0) log_write_end();

#if 0
    for(rep=0; rep<warm_rep; ++rep) {
      linpack(size);
    }

    /* Output cool-down marker. */
    barrier();
    if (my_rank == 0) {
      log_write_marker("run_end");
    }
    barrier();

    /* Write idle shoulder. */
    nanosleep(idle_length); 
    barrier();
    if (my_rank == 0) log_write_marker("idle_end");
#endif

    coherence_flush_all();
    barrier();
    
    if (my_rank == 0) {
      
      ret = verify();
  
      sol_t = (double)(end_t - start_t)/1E9;
      sol_acc = (double)(acc_t)/1E9;
      sol_ops = 2.0*n*n*n/3.0 + 2.0*n*n;
      
      log_write_result_double("flop_count", sol_ops*(double)nrep);
      log_write_result_double("raw_time", sol_t);
      log_write_result_double("time", sol_acc);

      printf("solution time %llu cycles = %g s exta %g s MFLOPS %g\n",
	     acc_t, 
	     sol_acc, 
	     (sol_t - sol_acc),
	     (sol_ops*nrep)/(sol_acc*1E6));
    } else {
      
      ret = 0;
      
    }

    barrier();
  }

  return ret;
}
