/*
  Simple barrier in multi-core shared memory.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include <string.h>
#include <stdint.h>
#include <c6x.h>

#include "include/barrier.h"
#include "include/cache.h"

#define FALSE 0
#define TRUE (!FALSE)

#define MAX_RANK 30

/*
  Counter with padding for 1 cache line (64 bytes).
*/
struct barrierCounter_t {
  volatile uint32_t count;
  uint32_t pad[31];
};

/*
  The actual barrier as it is laid out in memory.
*/
struct barrier_t {
  struct barrierCounter_t acknowledge;
  struct barrierCounter_t requests[MAX_RANK + 1];
};

/*
  Put barrier at end of MSMC memory.
*/
static struct barrier_t * const barrierData =
  (struct barrier_t *)(0x0c400000 - sizeof(struct barrier_t));


void 
barrier()
{
  unsigned numRanks, myRank, i;
  uint32_t round;
  /* Figure out who we are! */
  myRank = cache_get_mm_id();
  numRanks = cache_get_mm_num();
 
  cache_l1d_inv(barrierData, sizeof(*barrierData), TRUE);
  round = barrierData->acknowledge.count + 1;
  barrierData->requests[myRank].count = round;
  
  for(i=0; i<numRanks; ++i) 
    if (barrierData->requests[i].count != round) break;

  if (i<numRanks) {		/* Not the last one need to wait! */
    while (1) {
      cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);
      if (barrierData->acknowledge.count == round) break;
    }
  } else {			/* Last one, inform the waiting cores */
    barrierData->acknowledge.count = round;
    mfence();
    cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);
  }
}

void 
barrier_init()
{
  int allDone;
  unsigned numRanks, myRank, i; 

  /* Figure out who we are! */
  myRank = cache_get_mm_id();
  numRanks = cache_get_mm_num();

  /* Flush out all stale data */
  cache_l1d_inv(barrierData, sizeof(*barrierData), TRUE);

  /* Zero-fill all the barrier counters. */
  memset(barrierData, 0, sizeof(*barrierData));
  barrierData->acknowledge.count = 1;
  cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);

  /* We know that the memory is now set to 0 
     even if it was not before this. */

  /* Spin on all counter while setting the own to 1. */
  while (1) {
    cache_l1d_inv(barrierData, sizeof(*barrierData), TRUE);
    barrierData->requests[myRank].count = 1;
    cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);

    /* Check that all the cores have started spinning on this one. */
    for(i=0; i<numRanks; ++i) 
      if (barrierData->requests[i].count < 1) break;

    /* In case we see a 1 in all ranks we are done for this time. */
    if (i>=numRanks) break;
  }

  /* We now know that somebody must have set each slot to 1 and 
     we assume it was the "right" core executing this loop too.
     That means all cores were spinning on their own counters here. */

  /* Now we simply do some barriers to catch any misbehaviour! */
  barrier();
  barrier();
  

#if 0
  barrierData->requests[myRank].count = 0;
  mfence();
  cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);

  if (myRank == 0) {		/* Leader waits for the pack! */
    allDone = 0;

    while (!allDone) {
      cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);
      for(i=0; i<numRanks; ++i) 
	if (barrierData->requests[i].count != 0) break;

      allDone = (i == numRanks);
    }

    cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);
    barrierData->acknowledge.count = 0;
    mfence();
    cache_l1d_wbinv(barrierData, sizeof(*barrierData), TRUE);
  } else {			/* Others wait for the leader */
    while (1) {	
      cache_l1d_inv(barrierData, sizeof(*barrierData), TRUE);
      if ((barrierData->requests[0].count <= 1) &&
	  (barrierData->acknowledge.count == 0)) break;
    }
  }
  
  /* Final barrier to be on the safe side. */
  barrier();
#endif
}

