/*
  HyperLink runtime support functions.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "include/hl.h"
#include "include/hl_regs.h"
#include "include/cl.h"
#include "include/cl_regs.h"
#include "include/time.h"

#include <errno.h>
#include <stdio.h>

/*
  HyperLink parameter table entry.
*/
struct hl_link_parameters 
{
  uint32_t pll_config;
  uint32_t serdes_rx_config;
  uint32_t serdes_tx_config;
  uint32_t lane_power_mgmt;
};

/*
  HyperLink parameter table (build-in so far).
*/
static const struct hl_link_parameters hl_link_parameters[]= {
  /* 10.0 GBps @4lanes, 312.5 MHz refclk */
  {
    0x041,		/* MPY=32 (x8), VRANGE=0, LOOPBW=00 (medium)*/
    0x0046c084,		/* ENOC=1, EQHLD=0, EQ=1, CDR=5, LOS=4, TERM=1
			   INVPAIR=0, RATE=0, BUSWIDTH=2, ENRX=0 */
    0x000d0d04,		/* LB=0, MSYNC=0,FIRUPT=1, TWPST1=20, TWPRE=1,
			   SWING=10, INVPAIR=0, RATE=0, BUSWIDTH=010, ENTX=0 */
    0x07075e0c	        /* H2L=7, L2H=7, PWC=0x5e, HS=1, QL=1, SL=0, ZL=0 */
  }
};

static const int NUM_LANES = 4;

/**
   Default transmit address overlay setting.

   No security bit, no priviledge id, 256 MB trim.
*/
static const uint32_t TX_ADDR_OVL_DEFAULT = 0x0000000bul;

/**
   Default recieve address selector control settings.
   
   Security bit fixed to 0, take from bit 31.
   PrivId fixed to 0
   Segment selector set to bit [27-22] offset 0x3fffff (rxsegsel=6)
*/
static const uint32_t RX_ADDR_SELECTOR_DEFAULT = 0x000f0006ul;

/**
   Size of a RX translation segment.
*/
static const uint32_t RX_ADDR_SEG_SIZE = 0x00400000ul;

/**
   Encoded RX translation segment length.
*/
static const uint32_t RX_ADDR_SEG_LEN = 21;
   

int 
hl_init(unsigned param_idx)
{
  const struct hl_link_parameters *link_param;
  int lane;
  unsigned privid;
  unsigned segidx;

  link_param = hl_link_parameters + param_idx;

  cl_kick_unlock();
  
  /* FIXME: Power up the module and take out of reset. */
  
  /* Set the Link into reset and then setup the SERDES registers. */
  hl_module_registers->local.control = 0x4; /* Serial Stop */
  nanosleep(10000);
  hl_module_registers->local.control = 0x5; /* Stop & Reset */
  nanosleep(10000);
  
  /* Set up the SERDES and PLL config. */
  cl_registers->hl_serdes_pll_config = link_param->pll_config;
  for(lane=0; lane<NUM_LANES; ++lane) {
    cl_registers->hl_serdes_config[lane].rx_config = link_param->serdes_rx_config;
    cl_registers->hl_serdes_config[lane].tx_config = link_param->serdes_tx_config;
  }
  hl_module_registers->local.control = 0x4; /* Take out of reset */

  /* Wait for PLL to lock. */
  while ((cl_registers->hl_serdes_pll_config & 0x1) != 0x1);
  while ((cl_registers->hl_serdes_status & 0x1) != 0x1);

  /* Check that the link is initialized. */
  while ((hl_module_registers->local.status & 0x1) != 0x1);

  /* Enable the serial logic. */
  hl_module_registers->local.control = 0x0; /* Allow serial operations */
  
  /* Init the outbound address translation logic. */
  hl_module_registers->local.tx_addr_overlay = TX_ADDR_OVL_DEFAULT;

  /* Init the inbound address translation logic. */
  hl_module_registers->local.rx_addr_selector = RX_ADDR_SELECTOR_DEFAULT;
  
  for(privid=0; privid<16; ++privid) {
    hl_module_registers->local.rx_addr_privid_idx = privid & 0xf;
    hl_module_registers->local.rx_addr_privid_val = 0;
  }
  hl_module_registers->local.rx_addr_privid_idx = 0;

  for(segidx=0; segidx<64; ++segidx) {
    hl_module_registers->local.rx_addr_seg_idx = segidx & 0x3f;
    hl_module_registers->local.rx_addr_seg_val = 0;
  }
  hl_module_registers->local.rx_addr_seg_idx = 0;

  cl_kick_lock();

  return 0;
}

uint32_t 
hl_get_revision(enum hl_location location)
{
  uint32_t rev_id;

  switch (location) {
  case HL_LOCATION_LOCAL: 
    rev_id = hl_module_registers->local.revision_id;
    break;

  case HL_LOCATION_REMOTE: 
    rev_id = hl_module_registers->remote.revision_id;
    break;

  default:
    rev_id = 0;
    errno = ERANGE;
    break;
  }

  return rev_id;
}


int 
hl_ingess_addr_map(uint32_t remote_start, 
		   void *local_start,
		   uint32_t size)
{
  const uint32_t rx_offset_mask = RX_ADDR_SEG_SIZE - 1;
  unsigned start_idx;
  unsigned end_idx;
  unsigned seg_idx;
  uint32_t local_addr;
  uint32_t seg_val;

  if (((remote_start & rx_offset_mask) != 0) ||
      (((uint32_t)local_start & rx_offset_mask) !=0) ||
      ((size & rx_offset_mask) != 0)) {
    errno = ERANGE;
    return ERANGE;
  }

  local_addr = (uint32_t)local_start;
  start_idx = remote_start / RX_ADDR_SEG_SIZE;
  end_idx = (remote_start + size) / RX_ADDR_SEG_SIZE;

  if ((start_idx > 63) || (end_idx > 64)) {
    errno = ERANGE;
    return ERANGE;
  }

  for(seg_idx=start_idx; 
      seg_idx<end_idx; 
      ++seg_idx, local_addr+=RX_ADDR_SEG_SIZE) {
    seg_val = (local_addr & 0xffff0000) | RX_ADDR_SEG_LEN;

    printf("HL: Setting mapping for %08x to %08x idx=%02x val=%08x\n",
	   remote_start, 
	   local_addr,
	   seg_idx,
	   seg_val);
    
    hl_module_registers->local.rx_addr_seg_idx = seg_idx;
    hl_module_registers->local.rx_addr_seg_val = seg_val;
      
  }
  hl_module_registers->local.rx_addr_seg_idx = 0;

  return 0;
}

