/*
  Simple cache control.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef RT_CACHE_H
#define RT_CACHE_H

#include <stddef.h>
#include <stdint.h>
#include <c6x.h>

#include "corepack_regs.h"

static const uint32_t CACHE_L2CFG_NUMMM_MASK = 0x0f000000ul;
static const unsigned CACHE_L2CFG_NUMMM_SHIFT = 24;
static const uint32_t CACHE_L2CFG_MMID_MASK = 0x000f0000ul;
static const unsigned CACHE_L2CFG_MMID_SHIFT = 16;

static const uint32_t CACHE_L2_LINE_SIZE = 128;
static const uint32_t CACHE_L1P_LINE_SIZE = 32;
static const uint32_t CACHE_L1D_LINE_SIZE = 64;

/**
   Wait for all memory operations to complete.
*/
static inline void mfence()
{
  _mfence();
}

static inline unsigned cache_get_mm_num (void)
{    
  uint32_t val;
  
  val =  corepack_regs->cache_regs.l2cfg;
  
  return (unsigned)((val & CACHE_L2CFG_NUMMM_MASK) >> CACHE_L2CFG_NUMMM_SHIFT)+1;
}

static inline unsigned cache_get_mm_id (void)
{    
  uint32_t val;
  
  val =  corepack_regs->cache_regs.l2cfg;
  
  return (unsigned)((val & CACHE_L2CFG_MMID_MASK) >> CACHE_L2CFG_MMID_SHIFT);
}

static inline void cache_l1p_inv(void *base,
				 size_t size,
				 int wait)
{
  mfence();
  corepack_regs->cache_regs.l1pibar = (uint32_t) base;
  corepack_regs->cache_regs.l1piwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}


static inline void cache_l1d_wb(void *base,
				size_t size,
				int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dwbar = (uint32_t) base;
  corepack_regs->cache_regs.l1dwwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l1d_inv(void *base,
				 size_t size,
				 int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dibar = (uint32_t) base;
  corepack_regs->cache_regs.l1diwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l1d_wbinv(void *base,
				   size_t size,
				   int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dwibar = (uint32_t) base;
  corepack_regs->cache_regs.l1dwiwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l1d_wb_all(int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dwb = 1;
  
  if (wait) {
    while ((corepack_regs->cache_regs.l1dwb & 1) != 0);
  }
}

static inline void cache_l1d_inv_all(int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dinv = 1;
  
  if (wait) {
    while ((corepack_regs->cache_regs.l1dinv & 1) != 0);
  }
}

static inline void cache_l1d_wbinv_all(int wait)
{
  mfence();
  corepack_regs->cache_regs.l1dwbinv = 1;
  
  if (wait) {
    while ((corepack_regs->cache_regs.l1dwbinv & 1) != 0);
  }
}

static inline void cache_l2_wb(void *base,
			       size_t size,
			       int wait)
{
  mfence();
  corepack_regs->cache_regs.l2wbar = (uint32_t) base;
  corepack_regs->cache_regs.l2wwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l2_inv(void *base,
				size_t size,
				int wait)
{
  mfence();
  corepack_regs->cache_regs.l2ibar = (uint32_t) base;
  corepack_regs->cache_regs.l2iwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l2_wbinv(void *base,
				  size_t size,
				  int wait)
{
  mfence();
  corepack_regs->cache_regs.l2wibar = (uint32_t) base;
  corepack_regs->cache_regs.l2wiwc = (uint32_t) (size/4);

  if (wait) {
    mfence();
  }
}

static inline void cache_l2_wb_all(int wait)
{
  mfence();
  corepack_regs->cache_regs.l2wb = 1;
  
  if (wait) {
    while ((corepack_regs->cache_regs.l2wb & 1) != 0);
  }
}

static inline void cache_l2_inv_all(int wait)
{
  mfence();
  corepack_regs->cache_regs.l2inv = 1;
  
  if (wait) {
    while ((corepack_regs->cache_regs.l2inv & 1) != 0);
  }
}

static inline void cache_l2_wbinv_all(int wait)
{
  unsigned int int_val;

  int_val = _disable_interrupts();

  mfence();
  corepack_regs->cache_regs.l2wbinv = 1;
  
  if (wait) {
    mfence();
    while ((corepack_regs->cache_regs.l2wbinv & 1) != 0);
    
    /* Issue 16 nops! */
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
    asm(" nop 1");
  }
  
  _restore_interrupts(int_val);
}


#define CACHE_MAR_CACHEABLE ((uint32_t)(0x1))
#define CACHE_MAR_PREFETCHABLE ((uint32_t)(0x8))
#define CACHE_MAR_SIZE ((size_t)(0x01000000ull))

unsigned cache_set_memory_attributes(void *base, 
				     size_t size, 
				     uint32_t attributes);
uint32_t cache_get_memory_attribtues(void *base);

unsigned cache_l2_set_size(uint32_t size_log_2);
unsigned cache_l1d_set_size(uint32_t size_log_2);

#endif
