/**
   EDMA3 functions
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef RT_EDMA3_H
#define RT_EDMA3_H

#include "edma3_regs.h"

/**
   Set the specified PaRAM entry to the NULL descriptor.
*/
static inline void
edma3_param_set_null(unsigned ctrl, unsigned entry)
{
  volatile struct edma3_param_set *e;
  
  e = edma3cc_registers[ctrl].param + entry;

  e->opt = 0;
  e->src = 0;
  e->abcnt = 0;
  e->dst = 0;
  e->bidx = 0;
  e->link_bcntrld = 0x0000fffful;
  e->cidx = 0;
  e->ccnt = 0;
}

/**
   Fill the PaRAM of the specified controller with NULL descriptors.
*/
static inline void
edma3_param_init(unsigned ctrl)
{
  unsigned num_entries, entry;

  switch(ctrl) {
  case 0: num_entries = 128; break;
  case 1: 
  case 2: num_entries = 512; break;
  default: num_entries = 0; break;
  }

  for(entry=0; entry<num_entries; ++entry)
    edma3_param_set_null(ctrl, entry);
}

/**
   Copy the PaRAM set entry from src to dst. 
*/
static inline void
edma3_param_copy(unsigned ctrl, unsigned dst, unsigned src)
{
  volatile uint32_t *srcp, *dstp;
  unsigned cnt;

  srcp = (uint32_t *)(edma3cc_registers[ctrl].param + src);
  dstp = (uint32_t *)(edma3cc_registers[ctrl].param + dst);

  for(cnt=0; cnt<(sizeof(struct edma3_param_set)/sizeof(uint32_t)); ++cnt)
    *dstp++ = *srcp++;
}

/**
   Associate the specified DMA channel with the PaRAM entry.
*/
static inline void
edma3_set_dma_channel(unsigned ctrl, unsigned channel, unsigned param_entry)
{
  edma3cc_registers[ctrl].dchmap[channel] = param_entry << 5;
}

/**
   Associate the specified QDMA channel with the PaRAM entry.
*/
static inline void
edma3_set_qdma_channel(unsigned ctrl, 
		       unsigned channel, 
		       unsigned param_entry,
		       unsigned trigger_word)
{
  edma3cc_registers[ctrl].qchmap[channel] = 
    (param_entry << 5) | (trigger_word << 2);
}

/**
   Associate the specified DMA channel with the given event queue.
*/
static inline void
edma3_set_dma_queue(unsigned ctrl, unsigned channel, unsigned queue)
{
  unsigned idx, ofs;
  uint32_t mask;
  unsigned val;

  idx = channel/8;
  ofs = (channel%8)*4;
  mask = 0x7 << ofs;
  
  val = edma3cc_registers[ctrl].dmaqnum[idx];
  val = val & ~mask | ((queue & 0x3) << ofs);
  edma3cc_registers[ctrl].dmaqnum[idx] = val;
}

static inline uint32_t
edma3_calc_abcnt(unsigned acnt, unsigned bcnt)
{
  uint32_t ret;
  ret = (acnt & 0xffff) | ((bcnt & 0xffff)<<16);
  return ret;
}

static inline uint32_t
edma3_calc_link(unsigned link, unsigned bcntrld)
{
  uint32_t ret;
  ret = (link & 0xffff) | ((bcntrld & 0xffff)<<16);
  return ret;
}

static inline uint32_t
edma3_calc_idx(int src, int dst)
{
  uint32_t ret;
  ret = (uint32_t)(src & 0xffff) | ((uint32_t)(dst & 0xffff)<<16);
  return ret;
}	       

/**
   EDMA3 PaRAM transfer options.
*/

static const uint32_t edma3_opt_syncdim = (1 << 2);
static const uint32_t edma3_opt_static = (1 << 3);
static const uint32_t edma3_opt_tccmode = (1 << 11);
static const uint32_t edma3_opt_tcinten = (1 << 20);
static const uint32_t edma3_opt_itcinten = (1 << 21);
static const uint32_t edma3_opt_tcchen = (1 << 22);
static const uint32_t edma3_opt_itcchen = (1 << 23);

static inline uint32_t 
edma3_calc_opt_tcc(unsigned tcc)
{
  return (tcc & 0x3f) << 12;
}



#endif
